import numpy as np
import waLBerla as wlb
import matplotlib.pyplot as plt
from waLBerla.core_extension import makeSlice
from raytracerTest import runRayTracer


class Scenario:
    def __init__(self, reynoldsNumber=10.0, pdfReconstructionModel='OnlyMissing'):
        self.scenario = 1
        #> Domain Parameters
        self.domain_x = 51
        self.domain_y = self.domain_x
        self.domain_z = self.domain_x
        self.beam_size_x = np.shape(np.arange(-self.domain_x // 4 + 1, self.domain_x // 4 + 1))[0]
        print(self.beam_size_x)
        self.beam_size_y = np.shape(np.arange(-self.domain_y // 4 + 1, self.domain_y // 4 + 1))[0]
        print(self.beam_size_y)
        self.beam_size = int(self.beam_size_x * self.beam_size_y)
        print(self.beam_size)
        self.blocks = (1, 1, 1)
        self.periodicity = (1, 1, 0)
        self.cells_per_block = (self.domain_x // self.blocks[0], self.domain_y // self.blocks[1], self.domain_z // self.blocks[2])
        self.reynoldsNumber = reynoldsNumber
        self.pdfReconstructionModel = pdfReconstructionModel
        self.initAmplitude = None  # This will be set later in the config method

        self.config_dict = self.config(print_dict=True)
        self.current_timestep = 0
        self.total_timesteps = self.config_dict['PhysicsParameters']['timesteps']
        # self.electrons_across_iterations = []
        self.electrons_across_timesteps = np.zeros((self.total_timesteps, 4))  # 50 timesteps, 4 detectors
        self.z_heights = np.zeros((self.total_timesteps, self.beam_size))  # 50 timesteps, 25 cells
        self.fill_level_b = np.zeros((self.total_timesteps, self.beam_size))  # 50 timesteps, 25 cells, 3 components
        self.fill_level_a = np.zeros((self.total_timesteps, self.beam_size))  # 50 timesteps, 25 cells, 3 components
        self.normal_interp = np.zeros((self.total_timesteps, self.beam_size, 3))  # 50 timesteps, 25 cells, 3 components
        self.normals_b = np.zeros((self.total_timesteps, self.beam_size, 3))  # 50 timesteps, 25 cells, 3 components
        self.normals_a = np.zeros((self.total_timesteps, self.beam_size, 3))  # 50 timesteps, 25 cells, 3 components
        # self.initAmplitude = self.config_dict['DomainParameters']['initialAmplitude']


    @wlb.member_callback
    def config(self, print_dict=False):
        print('config_dict')
        # Set total_timesteps before using it
        self.total_timesteps = 864  # 864 // 16  # This should match the value in PhysicsParameters
        config_dict = dict(
            BlockForestParameters = dict(
                cellsPerBlock = self.cells_per_block,
                periodicity = self.periodicity,
                loadBalancingFrequency = 0,
                printLoadBalancingStatistics = False
            ),
            DomainParameters = dict(
                domainWidth = self.domain_x,  # equivalent to wavelength
                domainY = self.domain_y,
                liquidDepth = self.domain_z // 2,
                initialAmplitude = self.domain_z // 4
            ),
            PhysicsParameters = dict(
                reynoldsNumber = self.reynoldsNumber,
                relaxationRate = 1.8,
                enableWetting = False,
                contactAngle = 0,  # only used if enableWetting=true
                timesteps = self.total_timesteps
            ),
            ModelParameters = dict(
                pdfReconstructionModel = self.pdfReconstructionModel,
                pdfRefillingModel = 'EquilibriumRefilling',
                excessMassDistributionModel = 'EvenlyAllInterface',  # WeightedAllInterface
                curvatureModel = 'FiniteDifferenceMethod',
                enableForceWeighting = False,
                useSimpleMassExchange = True,
                cellConversionThreshold = 0.0,
                cellConversionForceThreshold = 0.0,
                enableBubbleModel = False,
                enableBubbleSplits = False  # only used if enableBubbleModel=true
            ),
            EvaluationParameters = dict(
                performanceLogFrequency = 1,
                evaluationFrequency = 1,
                filename = f'gravity_wave_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.txt'
            ),
            BoundaryParameters = dict(
                Border = [
                    # X
                    # dict(direction = 'W', walldistance = -1, flag = 'NoSlip'),
                    # dict(direction = 'E', walldistance = -1, flag = 'NoSlip'),
                    # Y
                    # dict(direction = 'N', walldistance = -1, flag = 'NoSlip'),
                    # dict(direction = 'S', walldistance = -1, flag = 'NoSlip'),
                    # Z
                    dict(direction = 'T', walldistance = -1, flag = 'NoSlip'),
                    dict(direction = 'B', walldistance = -1, flag = 'NoSlip')
                ]
            ),
            MeshOutputParameters = dict(
                writeFrequency = 864,
                baseFolder = f'mesh_out_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}'
            ),
            VTK = dict(
                fluid_field = dict(
                    writeFrequency = 1,
                    ghostLayers = 0,
                    baseFolder = f'vtk_out_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}',
                    samplingResolution = 1,
                    writers = dict(
                        fill_level = True,
                        mapped_flag = True,
                        velocity = True,
                        density = True,
                        curvature = False,
                        normal = False,
                        obstacle_normal = False,
                        pdf = False,
                        flag = False,
                        force_density = True,
                    ),
                    inclusion_filters = dict(
                        # liquidInterfaceFilter = False, # only needed if writing the ghost layer
                    ),
                    before_functions = dict(
                        ghost_layer_synchronization = False, # only needed if writing the ghost layer
                        gas_cell_zero_setter = True, # sets velocity=0 and density=1 all gas cells before writing VTK output
                    )
                ),
                domain_decomposition = dict(
                    writeFrequency = 1,
                    baseFolder = f'vtk_out_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}',
                    outputDomainDecomposition = True
                )
            )
        )

        if print_dict:
            print(config_dict)

        # Update class attributes
        self.initAmplitude = config_dict['DomainParameters']['initialAmplitude']
        # self.total_timesteps = config_dict['PhysicsParameters']['timesteps']
        return config_dict


    @wlb.member_callback
    def results_callback(self, **kwargs):
        print('executing results_callback')
        # print('current timestep (GravityWaveConfig.py):', self.current_timestep)
        # print("kwargs: ", kwargs)
        # fillFieldID = kwargs["fillFieldID"]
        # print('fillField', fillFieldID)
        fill_level_array = np.array([*kwargs.values()])
        # print(fill_level_array)
        electrons_per_detector_per_step, z_heights, fill_level_b, fill_level_a, normal_interp, normals_b, normals_a = runRayTracer(self.domain_x, self.domain_y, self.domain_z, fill_level_array, self.current_timestep)
        # print('finished raytracing at timestep:', self.current_timestep)
        # print('electrons_per_detector_per_step:', electrons_per_detector_per_step)

        # Sum electrons across all domain points for each detector
        total_electrons_for_detectors = np.sum(electrons_per_detector_per_step, axis=(0, 1))
        # print('total_electrons_for_detectors:', total_electrons_for_detectors)

        # Store the result for this timestep
        if self.current_timestep < self.total_timesteps:
            self.electrons_across_timesteps[self.current_timestep] = total_electrons_for_detectors

            self.z_heights[self.current_timestep] = z_heights
            self.normal_interp[self.current_timestep] = normal_interp
            self.normals_b[self.current_timestep] = normals_b
            self.normals_a[self.current_timestep] = normals_a
            self.fill_level_b[self.current_timestep] = fill_level_b
            self.fill_level_a[self.current_timestep] = fill_level_a

            '''print('electrons_across_timesteps:', self.electrons_across_timesteps[self.current_timestep])
            print('z_heights:', self.z_heights[self.current_timestep])
            print('normal_interp:', self.normal_interp[self.current_timestep])
            print('normals_b:', self.normals_b[self.current_timestep])
            print('normals_a:', self.normals_a[self.current_timestep])
            print('fill_level_b:', self.fill_level_b[self.current_timestep])
            print('fill_level_a:', self.fill_level_a[self.current_timestep])'''

            self.current_timestep += 1

        # print('electrons_across_timesteps:', self.electrons_across_timesteps)

        if self.current_timestep == self.total_timesteps:
            self.plot_electron_capture1()
            self.plot_z_heights_cell_2_0()
            # self.plot_fill_level_b()
            # self.plot_fill_level_a()
            # self.plot_normal_interp_x()
            # self.plot_normals_b_x()
            # self.plot_normals_a_x()
            self.plot_combined_results_cell_2_0()


    def plot_electron_capture(self):
        print('executing plot_electron_capture')

        # Convert list to numpy array if it's not already
        electrons_array = np.array(self.electrons_across_timesteps)

        plt.figure(figsize=(12, 8))
        timesteps = range(0, len(electrons_array))

        markers = ['o', 's', '^', 'D']  # Different markers for each detector

        for detector in range(4):
            plt.plot(timesteps, electrons_array[:, detector],
                     label=f'Detector {detector + 1}',
                     marker=markers[detector],
                     markevery=max(1, len(timesteps) // 20),  # Show about 20 markers along the line
                     markersize=6)

        plt.title("Electrons Captured by Detectors Across Timesteps")
        plt.xlabel("Timestep")
        plt.ylabel("Electrons Captured")
        plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/electrons_captured_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()

        # Print some debug information
        print(f"Shape of electrons_array: {electrons_array.shape}")
        print(f"Min value: {np.min(electrons_array)}")
        print(f"Max value: {np.max(electrons_array)}")


    def plot_electron_capture1(self):
        print('executing plot_electron_capture')

        # Convert list to numpy array if it's not already
        electrons_array = np.array(self.electrons_across_timesteps)

        timesteps = range(0, len(electrons_array))
        markers = ['o', 's', '^', 'D']  # Different markers for each detector

        # Plot for Detectors 1 and 2
        plt.figure(figsize=(12, 8))
        for detector in range(2):
            plt.plot(timesteps, electrons_array[:, detector],
                     label=f'Detector {detector + 1}',
                     marker=markers[detector],
                     markevery=20,  # max(1, 50),
                     markersize=6)

        plt.title("Electrons Captured by Detectors 1 and 2 Across Timesteps")
        plt.xlabel("Timestep")
        plt.ylabel("Electrons Captured")
        plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/electrons_captured_12_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()

        # Plot for Detectors 3 and 4
        plt.figure(figsize=(12, 8))
        for detector in range(2, 4):
            plt.plot(timesteps, electrons_array[:, detector],
                     label=f'Detector {detector + 1}',
                     marker=markers[detector],
                     markevery=20,  # max(1, 50),
                     markersize=6)

        plt.title("Electrons Captured by Detectors 3 and 4 Across Timesteps")
        plt.xlabel("Timestep")
        plt.ylabel("Electrons Captured")
        plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/electrons_captured_34_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()


    def plot_electron_capture2(self):
        print('executing plot_electron_capture')

        # Convert list to numpy array if it's not already
        electrons_array = np.array(self.electrons_across_timesteps)

        # Define timesteps
        timesteps = range(0, len(electrons_array))

        # Markers for each plot
        markers = ['o', 's', '^', 'D']  # Different markers for each detector

        # Loop over each detector and create a separate plot
        for detector in range(electrons_array.shape[1]):  # Assuming detectors are in columns
            plt.figure(figsize=(12, 8))
            plt.plot(timesteps, electrons_array[:, detector],
                     label=f'Detector {detector + 1}',
                     marker=markers[detector % len(markers)],  # Cycle through markers
                     markevery=2,  # Adjust marker frequency
                     markersize=6)

            # Set plot titles and labels
            plt.title(f"Electrons Captured by Detector {detector + 1} Across Timesteps")
            plt.xlabel("Timestep")
            plt.ylabel("Electrons Captured")
            plt.legend()
            plt.grid(True)

            # Adjust layout and save the figure
            plt.tight_layout()
            plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/electrons_captured_D{detector + 1}_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
            plt.close()


    def plot_z_heights(self):
        print('executing plot_z_heights')
        plt.figure(figsize=(12, 8))

        markers = ['o', 's', '^', 'D', 'v']  # Different markers
        timesteps = range(0, self.total_timesteps)
        for i, cell in enumerate(range(self.beam_size)):
            print(cell)
            z_values = self.z_heights[:, cell]  # Ensure z_values is defined here
            plt.plot(
                timesteps,
                z_values,
                label=f'Cell {cell + 1}',
                marker=markers[i % len(markers)],
                markevery=self.total_timesteps // 5,  # Add markers every 50 points
                linestyle='-' if i % 2 == 0 else '--'  # Alternate line styles
            )

            # Annotate every 50th point
            for j in range(0, self.total_timesteps, 50):
                plt.annotate(
                    f'{z_values[j]:.4f}',
                    (j + 1, float(z_values[j])),  # Ensure it's a tuple of floats
                    textcoords="offset points",
                    xytext=(5, -10),
                    ha='center',
                    fontsize=10
                )

                # Stagger annotation placement
                offset = (i % 5) * 50 - 0  # Adjust offset for staggering
                # Add annotation for cell index at the midpoint
                mid_timestep = min(self.total_timesteps // 2 + offset, len(z_values) - 1)
                plt.annotate(
                    f'Cell {cell + 1}',
                    (mid_timestep, float(z_values[mid_timestep])),
                    textcoords="offset points",
                    xytext=(0, 10),
                    ha='center',
                    fontsize=10
                )

        plt.title("Z-Heights Across Timesteps")
        plt.xlabel("Timestep")
        plt.ylabel("Z-Height")
        # plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/z_height_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()

        # Print some debug information
        print(f"Shape of z_heights array: {self.z_heights.shape}")
        print(f"Min value: {np.min(self.z_heights)}")
        print(f"Max value: {np.max(self.z_heights)}")


    def plot_z_heights1(self):
        print('executing plot_z_heights')
        plt.figure(figsize=(12, 8))

        markers = ['o', 's', '^', 'D', 'v']
        timesteps = range(0, self.total_timesteps)

        for i, cell in enumerate(range(self.beam_size)):
            z_values = self.z_heights[:, cell]
            plt.plot(
                timesteps,
                z_values,
                label=f'Cell {cell + 1}',
                marker=markers[i % len(markers)],
                markevery=self.total_timesteps // 5,
                linestyle='-' if i % 2 == 0 else '--'
            )

            # Annotate only the start, middle, and end
            for j in [0, self.total_timesteps // 2, self.total_timesteps - 1]:
                plt.annotate(
                    f'{z_values[j]:.4f}',
                    (j + 1, float(z_values[j])),
                    textcoords="offset points",
                    xytext=(5, -10),
                    ha='center',
                    fontsize=10
                )

        plt.title("Z-Heights Across Timesteps")
        plt.xlabel("Timestep")
        plt.ylabel("Z-Height")
        # plt.legend(loc='upper right', ncol=2)
        plt.grid(True)
        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/z_height_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()

        # Print some debug information
        print(f"Shape of z_heights array: {self.z_heights.shape}")
        print(f"Min value: {np.min(self.z_heights)}")
        print(f"Max value: {np.max(self.z_heights)}")


    def plot_z_heights_with_staggered_annotations(self):
        print('executing plot_z_heights_with_staggered_annotations')
        plt.figure(figsize=(12, 8))

        markers = ['o', 's', '^', 'D', 'v']
        timesteps = range(0, self.total_timesteps)
        offsets = np.linspace(-0.001, 0.001, self.beam_size)  # Calculate offsets for 25 cells

        for i, cell in enumerate(range(self.beam_size)):
            z_values = self.z_heights[:, cell]
            plt.plot(
                timesteps,
                z_values,
                label=f'Cell {cell + 1}',
                marker=markers[i % len(markers)],
                markevery=self.total_timesteps // 5,
                linestyle='-' if i % 2 == 0 else '--'
            )

            # Annotate every 50th point
            for j in range(0, self.total_timesteps, 50):
                plt.annotate(
                    f'{z_values[j]:.4f}',
                    (j + 1, float(z_values[j])),  # Ensure it's a tuple of floats
                    textcoords="offset points",
                    xytext=(5, -10),
                    ha='center',
                    fontsize=10
                )

                # Annotate only the midpoint for cell index
                mid_timestep = min(self.total_timesteps // 2, len(z_values) - 1)
                plt.annotate(
                    f'Cell {cell + 1}',
                    (mid_timestep, float(z_values[mid_timestep] + offsets[i])),  # Add offset
                    textcoords="offset points",
                    xytext=(0, 6),
                    ha='center',
                    fontsize=10
                )

        # Set y-axis limits
        z_min = np.min(self.z_heights)
        z_max = np.max(self.z_heights)
        plt.ylim(z_min - 0.0005, z_max + 0.0005)

        plt.title("Z-Heights Across Timesteps")
        plt.xlabel("Timestep")
        plt.ylabel("Z-Height")
        # plt.legend(loc='upper right')
        plt.grid(True)
        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/z_height_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()

        # Print some debug information
        print(f"Shape of z_heights array: {self.z_heights.shape}")
        print(f"Min value: {np.min(self.z_heights)}")
        print(f"Max value: {np.max(self.z_heights)}")


    def plot_z_heights_cell_2_0(self):
        """Plot z-height evolution over time for the specific cell position (2, 0)."""
        print('executing plot_z_heights for cell position (2, 0)')
        plt.figure(figsize=(12, 8))

        # Get z-values for the specific cell
        cell_index = 0  # Since you're only tracking one cell
        z_values = self.z_heights[:, cell_index]

        # Plot the z-height evolution
        timesteps = range(0, self.total_timesteps)
        plt.plot(
            timesteps,
            z_values,
            label='Cell (2, 0)',
            marker='o',
            markevery=10,  # self.total_timesteps // 10,  # Add markers at regular intervals
            linestyle='-'
        )

        # Annotate points
        for j in range(0, self.total_timesteps, max(1, self.total_timesteps // 10)):
            plt.annotate(
                f'{z_values[j]:.4f}',
                (j, z_values[j]),
                textcoords="offset points",
                xytext=(0, 10),
                ha='center',
                fontsize=8,
            )

        plt.title("Z-Height Evolution at Cell Position (2, 0)")
        plt.xlabel("Timestep")
        plt.ylabel("Z-Height (m)")
        plt.grid(True)
        plt.tight_layout()

        # Save the plot
        plt.savefig(
            f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/z_height_cell_2_0_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png",
            dpi=300
        )
        plt.close()

        # Print debug information
        print(f"Z-heights for cell (2, 0):")
        print(f"Min value: {np.min(z_values)}")
        print(f"Max value: {np.max(z_values)}")
        print(f"Mean value: {np.mean(z_values)}")

    def plot_fill_level_b(self):
        print('executing plot_fill_level_b')
        plt.figure(figsize=(12, 8))

        markers = ['o', 's', '^', 'D', 'v']
        timesteps = range(0, self.total_timesteps)
        # offsets = np.linspace(-0.0008, 0.0008, 25)  # Calculate offsets for 25 cells

        for i, cell in enumerate(range(self.beam_size)):
            fill_level_b = self.fill_level_b[:, cell]
            plt.plot(
                timesteps,
                fill_level_b,
                label=f'Cell {cell + 1}',
                marker=markers[i % len(markers)],
                markevery=self.total_timesteps // 5,
                linestyle='-' if i % 2 == 0 else '--'
            )

            # Annotate every 50th point with cell index
            for j in range(0, self.total_timesteps, 50):
                plt.annotate(
                    f'{fill_level_b[j]:.4f}',
                    (j + 1, float(fill_level_b[j])),
                    textcoords="offset points",
                    xytext=(5, -10),
                    ha='center',
                    fontsize=10
                )

                offset = (i % 5) * 50 - 0
                # Stagger annotation placement
                # Add annotation for cell index at the midpoint
                mid_timestep = min(self.total_timesteps // 2 + offset, len(fill_level_b) - 1)
                plt.annotate(
                    f'Cell {cell + 1}',
                    (mid_timestep, float(fill_level_b[mid_timestep])),  # Add offset
                    textcoords="offset points",
                    xytext=(0, 5),
                    ha='center',
                    fontsize=10
                )

        # Set y-axis limits
        n_min = np.min(self.fill_level_b)
        n_max = np.max(self.fill_level_b)
        plt.ylim(n_min - 0.1, n_max + 0.1)

        plt.title("Fill Levels Across Timesteps")
        plt.xlabel("Timestep")
        plt.ylabel("Fill Level Component")
        # plt.legend(loc='upper right')
        plt.grid(True)
        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/fill_level_b_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()

        # Print some debug information
        print(f"Shape of fill_level_b array: {self.fill_level_b.shape}")
        print(f"Min value: {np.min(self.fill_level_b)}")
        print(f"Max value: {np.max(self.fill_level_b)}")


    def plot_fill_level_a(self):
        print('executing plot_fill_level_a')
        plt.figure(figsize=(12, 8))

        markers = ['o', 's', '^', 'D', 'v']
        timesteps = range(0, self.total_timesteps)
        # offsets = np.linspace(-0.0008, 0.0008, 25)  # Calculate offsets for 25 cells

        for i, cell in enumerate(range(self.beam_size)):
            fill_level_a = self.fill_level_a[:, cell]
            plt.plot(
                timesteps,
                fill_level_a,
                label=f'Cell {cell + 1}',
                marker=markers[i % len(markers)],
                markevery=self.total_timesteps // 5,
                linestyle='-' if i % 2 == 0 else '--'
            )

            # Annotate every 50th point with cell index
            for j in range(0, self.total_timesteps, 50):
                plt.annotate(
                    f'{fill_level_a[j]:.4f}',
                    (j + 1, float(fill_level_a[j])),
                    textcoords="offset points",
                    xytext=(5, -10),
                    ha='center',
                    fontsize=10
                )

                offset = (i % 5) * 50 - 0  # Adjust offset for staggering
                # Stagger annotation placement
                # Add annotation for cell index at the midpoint
                mid_timestep = min(self.total_timesteps // 2 + offset, len(fill_level_a) - 1)
                plt.annotate(
                    f'Cell {cell + 1}',
                    (mid_timestep, float(fill_level_a[mid_timestep])),  # Add offset
                    textcoords="offset points",
                    xytext=(0, 5),
                    ha='center',
                    fontsize=10
                )

        # Set y-axis limits
        n_min = np.min(self.fill_level_a)
        n_max = np.max(self.fill_level_a)
        plt.ylim(n_min - 0.1, n_max + 0.1)

        plt.title("Fill Levels Across Timesteps")
        plt.xlabel("Timestep")
        plt.ylabel("Fill Level Component")
        # plt.legend(loc='upper right')
        plt.grid(True)
        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/fill_level_a_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()

        # Print some debug information
        print(f"Shape of fill_level_a array: {self.fill_level_a.shape}")
        print(f"Min value: {np.min(self.fill_level_a)}")
        print(f"Max value: {np.max(self.fill_level_a)}")


    def plot_normals(self):
        print('executing plot_normals')
        fig, axs = plt.subplots(1, 1, figsize=(12, 24))
        timesteps = range(0, self.total_timesteps)

        for cell in range(self.beam_size):
            axs[0].plot(timesteps, self.normal_interp[:, cell, 0], label=f'Cell {cell + 1}')
            axs[1].plot(timesteps, self.normal_interp[:, cell, 1], label=f'Cell {cell + 1}')
            axs[2].plot(timesteps, self.normal_interp[:, cell, 2], label=f'Cell {cell + 1}')

        axs[0].set_title("Normal X Component Across Timesteps")
        axs[1].set_title("Normal Y Component Across Timesteps")
        axs[2].set_title("Normal Z Component Across Timesteps")

        for ax in axs:
            ax.set_xlabel("Timestep")
            ax.set_ylabel("Normal Component")
            ax.legend()
            ax.grid(True)

        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/normals_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()

        # Print some debug information
        print(f"Shape of normals array: {self.normal_interp.shape}")
        print(f"Min value: {np.min(self.normal_interp)}")
        print(f"Max value: {np.max(self.normal_interp)}")


    def plot_normal_interp_x(self):
        print('executing plot_normal_interp_x')
        plt.figure(figsize=(12, 8))

        # for cell in range(self.beam_size):
        # plt.plot(range(1, self.total_timesteps + 1), self.normals[:, cell, 0], label=f'Cell {cell + 1}')

        markers = ['o', 's', '^', 'D', 'v']
        # offsets = np.linspace(-0.0008, 0.0008, 25)  # Calculate offsets for 25 cells

        for i, cell in enumerate(range(self.beam_size)):
            normal_interp_x = self.normal_interp[:, cell, 0]  # X component of normals
            plt.plot(
                range(0, self.total_timesteps),
                normal_interp_x,
                label=f'Cell {cell + 1}',
                marker=markers[i % len(markers)],
                markevery=self.total_timesteps // 5,
                linestyle='-' if i % 2 == 0 else '--'
            )

            # Annotate every 50th point with cell index
            for j in range(0, self.total_timesteps, 50):
                plt.annotate(
                    f'{normal_interp_x[j]:.4f}',
                    (j + 1, float(normal_interp_x[j])),
                    textcoords="offset points",
                    xytext=(5, -10),
                    ha='center',
                    fontsize=10
                )

                offset = (i % 5) * 50 - 0  # Adjust offset for staggering
                # Stagger annotation placement
                # Add annotation for cell index at the midpoint
                mid_timestep = min(self.total_timesteps // 2 + offset, len(normal_interp_x) - 1)
                plt.annotate(
                    f'Cell {cell + 1}',
                    (mid_timestep, float(normal_interp_x[mid_timestep])),  # Add offset
                    textcoords="offset points",
                    xytext=(0, 5),
                    ha='center',
                    fontsize=10
                )

        plt.title("Interpolated Normal X Component Across Timesteps")
        plt.xlabel("Timestep")
        plt.ylabel("Interpolated Normal X Component")
        # plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/normal_interp_x_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()

        # Print some debug information
        print(f"Shape of normal_interp array: {self.normal_interp.shape}")
        print(f"Min value of X component: {np.min(self.normal_interp[:, :, 0])}")
        print(f"Max value of X component: {np.max(self.normal_interp[:, :, 0])}")


    def plot_normals_b_x(self):
        print('executing normals_b_x')
        plt.figure(figsize=(12, 8))

        markers = ['o', 's', '^', 'D', 'v']
        timesteps = range(0, self.total_timesteps)

        for i, cell in enumerate(range(self.beam_size)):
            normals_b_x = self.normals_b[:, cell, 0]
            plt.plot(
                timesteps,
                normals_b_x,
                label=f'Cell {cell + 1}',
                marker=markers[i % len(markers)],
                markevery=self.total_timesteps // 5,
                linestyle='-' if i % 2 == 0 else '--'
            )

            # Annotate every 50th point with cell index
            for j in range(0, self.total_timesteps, 50):
                plt.annotate(
                    f'{normals_b_x[j]:.4f}',
                    (j + 1, float(normals_b_x[j])),
                    textcoords="offset points",
                    xytext=(5, -10),
                    ha='center',
                    fontsize=10
                )

                # Stagger annotation placement
                offset = (i % 5) * 50 - 0  # Adjust offset for staggering
                # Add annotation for cell index at the midpoint
                mid_timestep = min(self.total_timesteps // 2 + offset, len(normals_b_x) - 1)
                plt.annotate(
                    f'Cell {cell + 1}',
                    (mid_timestep, float(normals_b_x[mid_timestep])),
                    textcoords="offset points",
                    xytext=(0, 5),
                    ha='center',
                    fontsize=10
                )

        # Set y-axis limits
        n_min = np.min(self.normals_b[:, :, 0])
        n_max = np.max(self.normals_b[:, :, 0])
        plt.ylim(n_min - 0.1, n_max + 0.1)

        plt.title("Surface Normals Across Timesteps")
        plt.xlabel("Timestep")
        plt.ylabel("Normal X Component")
        # plt.legend(loc='upper right')
        plt.grid(True)
        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/normals_b_x_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()

        # Print some debug information
        print(f"Shape of normals_b array: {self.normals_b.shape}")
        print(f"Min value: {np.min(self.normals_b[:, :, 0])}")
        print(f"Max value: {np.max(self.normals_b[:, :, 0])}")


    def plot_normals_a_x(self):
        print('executing normals_a_x')
        plt.figure(figsize=(12, 8))

        markers = ['o', 's', '^', 'D', 'v']
        timesteps = range(0, self.total_timesteps)

        for i, cell in enumerate(range(self.beam_size)):
            normals_a_x = self.normals_a[:, cell, 0]
            plt.plot(
                timesteps,
                normals_a_x,
                label=f'Cell {cell + 1}',
                marker=markers[i % len(markers)],
                markevery=self.total_timesteps // 5,
                linestyle='-' if i % 2 == 0 else '--'
            )

            offset = (i % 5) * 50 - 0  # Adjust offset for staggering
            # Annotate at a specific point with staggered y-offsets
            mid_timestep = min(self.total_timesteps // 2 + offset, len(normals_a_x) - 1)
            plt.annotate(
                f'Cell {cell + 1}',
                (mid_timestep, float(normals_a_x[mid_timestep])),
                textcoords="offset points",
                xytext=(0, 5),
                ha='center',
                fontsize=10
            )

        # Set y-axis limits
        n_min = np.min(self.normals_a[:, :, 0])
        n_max = np.max(self.normals_a[:, :, 0])
        plt.ylim(n_min - 0.1, n_max + 0.1)

        plt.title("Surface Normals Across Timesteps")
        plt.xlabel("Timestep")
        plt.ylabel("Normal X Component")
        # plt.legend(loc='upper right')
        plt.grid(True)
        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/normals_a_x_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()

        # Print some debug information
        print(f"Shape of normals_b array: {self.normals_a.shape}")
        print(f"Min value: {np.min(self.normals_a[:, :, 0])}")
        print(f"Max value: {np.max(self.normals_a[:, :, 0])}")


    def plot_combined_results1(self):
        print('executing plot_combined_results1')
        # center_cells = [10, 11, 12, 13, 14]  # Indices of center cells (0-based)
        center_cells = [5, 6, 7, 8, 9]  # Indices of center cells (0-based)

        fig, axs = plt.subplots(7, 1, figsize=(12, 24), sharex=True)
        for ax in axs:
            ax.xaxis.set_tick_params(labelbottom=True)

        # plt.subplots_adjust(hspace=0.3)

        # Plot electron capture
        electrons_array = np.array(self.electrons_across_timesteps)
        timesteps = range(0, len(electrons_array))
        for detector in range(4):
            axs[0].plot(timesteps, electrons_array[:, detector], label=f'Detector {detector + 1}')
        axs[0].set_title("Electrons Captured by Detectors")
        axs[0].set_ylabel("Electrons Captured")
        axs[0].set_xlabel("Timestep")
        axs[0].legend()
        axs[0].grid(True)

        # Plot z-heights
        markers = ['o', 's', '^', 'D', 'v']
        offsets = np.linspace(-0.001, 0.001, len(center_cells))
        for i, cell in enumerate(center_cells):
            z_values = self.z_heights[:, cell]
            axs[1].plot(timesteps, z_values, label=f'Cell {cell + 1}', marker=markers[i], markevery=self.total_timesteps // 5, linestyle='-')
            mid_timestep = self.total_timesteps // 10
            axs[1].annotate(f'Cell {cell + 1}', (mid_timestep, z_values[mid_timestep] + offsets[i]), textcoords="offset points", xytext=(0, 6), ha='center', fontsize=10)
        axs[1].set_title("Z-Heights")
        axs[1].set_ylabel("Z-Height")
        axs[1].grid(True)

        # Plot fill level b
        for i, cell in enumerate(center_cells):
            fill_level_b = self.fill_level_b[:, cell]
            axs[2].plot(timesteps, fill_level_b, label=f'Cell {cell + 1}', marker=markers[i], markevery=self.total_timesteps // 5, linestyle='-')
            mid_timestep = self.total_timesteps // 10
            axs[2].annotate(f'Cell {cell + 1}', (mid_timestep, fill_level_b[mid_timestep]), textcoords="offset points", xytext=(0, 5), ha='center', fontsize=10)
        axs[2].set_title("Fill Levels (Below)")
        axs[2].set_ylabel("Fill Level")
        axs[2].grid(True)

        # Plot fill level a
        for i, cell in enumerate(center_cells):
            fill_level_a = self.fill_level_a[:, cell]
            axs[3].plot(timesteps, fill_level_a, label=f'Cell {cell + 1}', marker=markers[i], markevery=self.total_timesteps // 5, linestyle='-')
            mid_timestep = self.total_timesteps // 10
            axs[3].annotate(f'Cell {cell + 1}', (mid_timestep, fill_level_a[mid_timestep]), textcoords="offset points", xytext=(0, 5), ha='center', fontsize=10)
        axs[3].set_title("Fill Levels (Above)")
        axs[3].set_ylabel("Fill Level")
        axs[3].grid(True)

        # Plot normal interp x
        for i, cell in enumerate(center_cells):
            normal_interp_x = self.normal_interp[:, cell, 0]  # X component of normals
            axs[4].plot(timesteps, normal_interp_x, label=f'Cell {cell + 1}', marker=markers[i], markevery=self.total_timesteps // 5, linestyle='-')
            mid_timestep = min(self.total_timesteps // 2, len(normal_interp_x) - 1)
            axs[4].annotate(f'Cell {cell + 1}', (mid_timestep, normal_interp_x[mid_timestep]), textcoords="offset points", xytext=(0, 5), ha='center', fontsize=10)
        axs[4].set_title("Interpolated Normal X Component")
        axs[4].set_ylabel("Normal X")
        axs[4].grid(True)

        # Plot normals b x
        for i, cell in enumerate(center_cells):
            normals_b_x = self.normals_b[:, cell, 0]
            axs[5].plot(timesteps, normals_b_x, label=f'Cell {cell + 1}', marker=markers[i], markevery=self.total_timesteps // 5, linestyle='-')
            mid_timestep = min(self.total_timesteps // 2, len(normals_b_x) - 1)
            axs[5].annotate(f'Cell {cell + 1}', (mid_timestep, normals_b_x[mid_timestep]), textcoords="offset points", xytext=(0, 5), ha='center', fontsize=10)
        axs[5].set_title("Surface Normals (Below) X Component")
        axs[5].set_ylabel("Normal X")
        axs[5].grid(True)

        # Plot normals a x
        for i, cell in enumerate(center_cells):
            normals_a_x = self.normals_a[:, cell, 0]
            axs[6].plot(timesteps, normals_a_x, label=f'Cell {cell + 1}', marker=markers[i], markevery=self.total_timesteps // 5, linestyle='-')
            mid_timestep = min(self.total_timesteps // 2, len(normals_a_x) - 1)
            axs[6].annotate(f'Cell {cell + 1}', (mid_timestep, normals_a_x[mid_timestep]), textcoords="offset points", xytext=(0, 5), ha='center', fontsize=10)
        axs[6].set_title("Surface Normals (Above) X Component")
        axs[6].set_ylabel("Normal X")
        axs[6].set_xlabel("Timestep")
        axs[6].grid(True)

        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/combined_results_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()


    def plot_combined_results2(self):
        print('executing plot_combined_results')
        center_cells = [10, 11, 12, 13, 14]
        edge_cells_1 = [0, 1, 2, 3, 4, 20, 21, 22, 23, 24]
        edge_cells_2 = [5, 6, 7, 8, 9, 15, 16, 17, 18, 19]

        fig, axs = plt.subplots(7, 3, figsize=(36, 24), sharex=False)
        markers = ['o', 's', '^', 'D', 'v', 'p', '*', 'h', 'x', '+']

        # Common function to set x-axis for all subplots
        def set_x_axis(ax):
            ax.set_xlabel("Timestep")
            ax.set_xlim(0, self.total_timesteps - 1)
            ax.set_xticks(range(0, self.total_timesteps, 5))  # Show every 5th timestep
            ax.tick_params(axis='x', rotation=45)

        # Common plotting function
        def plot_data(ax, data, cells, title, ylabel):
            timesteps = range(0, len(data))
            for i, cell in enumerate(cells):
                ax.plot(timesteps, data[:, cell], label=f'Cell {cell + 1}',
                        marker=markers[i % len(markers)], markevery=self.total_timesteps // 5)
            ax.set_title(title)
            ax.set_ylabel(ylabel)
            ax.grid(True)
            ax.set_xlabel("Timestep")
            # if ax.get_subplotspec().is_last_row():
                # ax.set_xlabel("Timestep")
            ax.legend(loc='upper right', ncol=2)
            set_x_axis(ax)

        # Plot electron capture (same for all columns)
        electrons_array = np.array(self.electrons_across_timesteps)
        timesteps = range(0, len(electrons_array))
        for col in range(3):
            for detector in range(4):
                axs[0, col].plot(timesteps, electrons_array[:, detector], label=f'Detector {detector + 1}',
                                 marker=markers[detector % len(markers)], markevery=self.total_timesteps // 5)
            axs[0, col].set_title("Electrons Captured by Detectors")
            axs[0, col].set_ylabel("Electrons Captured")
            axs[0, col].legend()
            axs[0, col].grid(True)
            axs[0, col].set_xlabel("Timestep")

        # Plot data for each group of cells
        for col, cells in enumerate([center_cells, edge_cells_1, edge_cells_2]):
            plot_data(axs[1, col], self.z_heights, cells, "Z-Heights", "Z-Height")
            plot_data(axs[2, col], self.fill_level_b, cells, "Fill Levels (Below)", "Fill Level")
            plot_data(axs[3, col], self.fill_level_a, cells, "Fill Levels (Above)", "Fill Level")
            plot_data(axs[4, col], self.normal_interp[:, :, 0], cells, "Interpolated Normal X Component", "Normal X")
            plot_data(axs[5, col], self.normals_b[:, :, 0], cells, "Surface Normals (Below) X Component", "Normal X")
            plot_data(axs[6, col], self.normals_a[:, :, 0], cells, "Surface Normals (Above) X Component", "Normal X")

        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/combined_results_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()


    def plot_combined_results3(self):
        print('executing plot_combined_results')
        cells_1 = [0, 1, 2, 3, 4,]
        cells_2 = [5, 6, 7, 8, 9,]
        cells_3 = [10, 11, 12, 13, 14,]
        cells_4 = [15, 16, 17, 18, 19,]
        cells_5 = [20, 21, 22, 23, 24,]


        fig, axs = plt.subplots(7, 5, figsize=(60, 24), sharex=False)
        markers = ['o', 's', '^', 'D', 'v', 'p', '*', 'h', 'x', '+']

        timesteps = range(0, self.total_timesteps)

        # Common function to set x-axis for all subplots
        def set_x_axis(ax):
            ax.set_xlabel("Timestep")
            ax.set_xlim(0, self.total_timesteps - 1)
            ax.set_xticks(range(0, self.total_timesteps, 5))  # Show every 5th timestep
            ax.tick_params(axis='x', rotation=0)

        # Common plotting function
        def plot_data(ax, data, cells, title, ylabel):
            for i, cell in enumerate(cells):
                ax.plot(timesteps, data[:, cell], label=f'Cell {cell + 1}',
                        marker=markers[i % len(markers)], markevery=self.total_timesteps // 5)
            ax.set_title(title)
            ax.set_ylabel(ylabel)
            ax.grid(True)
            ax.set_xlabel("Timestep")
            ax.legend()
            set_x_axis(ax)

        # Plot data for each group of cells
        electrons_array = np.array(self.electrons_across_timesteps)

        for col, cells in enumerate([cells_1, cells_2, cells_3, cells_4, cells_5]):
            # Plot electron capture
            for detector in range(4):
                axs[0, col].plot(timesteps, electrons_array[:, detector], label=f'Detector {detector + 1}',
                                 marker=markers[detector % len(markers)], markevery=self.total_timesteps // 5)
            axs[0, col].set_title("Electrons Captured by Detectors")
            axs[0, col].set_ylabel("Electrons Captured")
            axs[0, col].legend()
            axs[0, col].grid(True)
            set_x_axis(axs[0, col])

            # Plot other data
            plot_data(axs[1, col], self.z_heights, cells, "Z-Heights", "Z-Height")
            plot_data(axs[2, col], self.fill_level_b, cells, "Fill Levels (Below)", "Fill Level")
            plot_data(axs[3, col], self.fill_level_a, cells, "Fill Levels (Above)", "Fill Level")
            plot_data(axs[4, col], self.normal_interp[:, :, 0], cells, "Interpolated Normal X Component", "Normal X")
            plot_data(axs[5, col], self.normals_b[:, :, 0], cells, "Surface Normals (Below) X Component", "Normal X")
            plot_data(axs[6, col], self.normals_a[:, :, 0], cells, "Surface Normals (Above) X Component", "Normal X")

        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/combined_results_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()


    def plot_combined_results_cell_2_0(self):
        print('executing plot_combined_results')
        # We're only interested in one cell (the one at position (2, 0))
        target_cell = 0  # Since we're only tracking one cell now

        fig, axs = plt.subplots(7, 1, figsize=(12, 24), sharex=True)
        for ax in axs:
            ax.xaxis.set_tick_params(labelbottom=True)

        # Plot timesteps
        timesteps = range(0, self.total_timesteps)

        # Plot electron capture
        electrons_array = np.array(self.electrons_across_timesteps)
        for detector in range(4):
            axs[0].plot(timesteps, electrons_array[:, detector],
                        label=f'Detector {detector + 1}')
        axs[0].set_title("Electrons Captured by Detectors")
        axs[0].set_ylabel("Electrons Captured")
        axs[0].set_xlabel("Timestep")
        axs[0].legend()
        axs[0].grid(True)

        # Plot z-heights for the target cell
        z_values = self.z_heights[:, target_cell]
        axs[1].plot(timesteps, z_values, label='Cell (2, 0)',
                    marker='o', markevery=20)
        axs[1].set_title("Z-Height at Cell Position (2, 0)")
        axs[1].set_ylabel("Z-Height")
        axs[1].grid(True)

        # Plot fill level below
        fill_level_b = self.fill_level_b[:, target_cell]
        axs[2].plot(timesteps, fill_level_b, label='Cell (2, 0)',
                    marker='o', markevery=20)
        axs[2].set_title("Fill Level (Below) at Cell Position (2, 0)")
        axs[2].set_ylabel("Fill Level")
        axs[2].grid(True)

        # Plot fill level above
        fill_level_a = self.fill_level_a[:, target_cell]
        axs[3].plot(timesteps, fill_level_a, label='Cell (2, 0)',
                    marker='o', markevery=20)
        axs[3].set_title("Fill Level (Above) at Cell Position (2, 0)")
        axs[3].set_ylabel("Fill Level")
        axs[3].grid(True)

        # Plot normal interp x component
        normal_interp_x = self.normal_interp[:, target_cell, 0]
        axs[4].plot(timesteps, normal_interp_x, label='Cell (2, 0)',
                    marker='o', markevery=20)
        axs[4].set_title("Interpolated Normal X Component at Cell Position (2, 0)")
        axs[4].set_ylabel("Normal X")
        axs[4].grid(True)

        # Plot normals below x component
        normals_b_x = self.normals_b[:, target_cell, 0]
        axs[5].plot(timesteps, normals_b_x, label='Cell (2, 0)',
                    marker='o', markevery=20)
        axs[5].set_title("Surface Normal (Below) X Component at Cell Position (2, 0)")
        axs[5].set_ylabel("Normal X")
        axs[5].grid(True)

        # Plot normals above x component
        normals_a_x = self.normals_a[:, target_cell, 0]
        axs[6].plot(timesteps, normals_a_x, label='Cell (2, 0)',
                    marker='o', markevery=20)
        axs[6].set_title("Surface Normal (Above) X Component at Cell Position (2, 0)")
        axs[6].set_ylabel("Normal X")
        axs[6].set_xlabel("Timestep")
        axs[6].grid(True)

        plt.tight_layout()
        plt.savefig(f"/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/combined_results_cell_2_0_{self.domain_x}_{self.domain_y}_{self.domain_z}_{self.initAmplitude}_{self.total_timesteps}.png", dpi=300)
        plt.close()
'''
    @wlb.member_callback
    def at_end_of_time_step(self, **kwargs):
        t = kwargs["timeStep"]
        # blockForest = kwargs["blockForest"]
        target_rank = kwargs["target_rank"]
        # print("blockForest: ", blockForest)
        print("kwargs: ", kwargs)
        wlb_field = False # wlb.field.gather(1, 'Fill level field', makeSlice[:, :, :], target_rank)
        if wlb_field:
            data = {'timestep': t}
            data.update(self.config_dict['BlockForestParameters'])
            data.update(self.config_dict['DomainParameters'])
            data.update(kwargs)

            fill_level_field = np.asarray(wlb_field).squeeze()
            assert np.isfinite(np.sum(fill_level_field)), "NaN detected in fill_level_field"
            runRayTracer(self.domain_x, self.domain_y, self.domain_z, fill_level_field)
'''

'''
@wlb.member_callback
def results_callback(self, **kwargs):
    cell_data = kwargs['cellData']
    fill_levels = np.array([data.fillLevel for data in cell_data])
    coordinates = np.array([[data.coordinates[0], data.coordinates[1], data.coordinates[2]] for data in cell_data])

    # Reshape fill_levels into a 3D array based on the domain size
    fill_levels_3d = fill_levels.reshape((self.domain_z + 1, self.domain_y + 1, self.domain_x + 1))

    # Pass the fill levels and coordinates to the raytracerTest.run function
    run(self.domain_x + 1, self.domain_y + 1, self.domain_z + 1, fill_levels_3d, coordinates)
'''


def gravity_wave():
    scenarios = wlb.ScenarioManager()
    print('scenarios = wlb.ScenarioManager()')
    scenario_instance = Scenario()
    print('scenario_instance = Scenario()')
    scenarios.add(scenario_instance)
    print('scenarios.add(scenario_instance)')

    # Run simulation here...
    # This should run for 50 timesteps

    # After all timesteps are complete, plot the results
    # scenario_instance.plot_electron_capture()
    # print('scenario_instance.plot_electron_capture()')


gravity_wave()
