import vtk
import os
from math import floor
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from matplotlib.animation import FuncAnimation
from typing import Tuple, List
import scipy.ndimage as ndimage
from scipy.ndimage import gaussian_filter
from scipy.spatial.transform import Rotation as Ro
from scipy.spatial.transform import Slerp
from scipy.interpolate import PchipInterpolator
import xml.etree.ElementTree as ET
from vtkmodules.util import numpy_support


def get_domain_dimensions(vti_file):
    """Get domain dimensions from VTI file"""
    reader = vtk.vtkXMLImageDataReader()
    reader.SetFileName(vti_file)
    reader.Update()
    data = reader.GetOutput()
    dims = data.GetDimensions()
    # For cell data, return dimensions without subtracting 1
    return dims[0]-1, dims[1]-1, dims[2]-1  # Convert from point dims to cell dims

def read_pvd_timesteps(_pvd_file):
    tree = ET.parse(_pvd_file)
    root = tree.getroot()
    timesteps = []
    filenames = []

    for dataset in root.findall(".//DataSet"):
        timesteps.append(float(dataset.get('timestep')))
        filenames.append(dataset.get('file'))

    return timesteps, filenames

'''def read_vti_data(file_path):
    reader = vtk.vtkXMLImageDataReader()
    reader.SetFileName(file_path)
    reader.Update()

    data = reader.GetOutput()
    point_data = data.GetPointData()

    fill_level = vtk.vtkFloatArray.SafeDownCast(point_data.GetArray("fill_level"))

    if fill_level is None:
        raise ValueError("fill_level data not found in the VTI file")

    np_array = vtk.util.numpy_support.vtk_to_numpy(fill_level)
    return np_array'''

def read_vti_data(file_path):
    """Read VTI file and extract fill level data"""
    reader = vtk.vtkXMLImageDataReader()
    reader.SetFileName(file_path)
    reader.Update()

    data = reader.GetOutput()
    cell_data = data.GetCellData()
    point_data = data.GetPointData()

    # Get dimensions
    dims = data.GetDimensions()
    cell_dims = (dims[0]-1, dims[1]-1, dims[2]-1)  # Convert to cell dimensions

    print(f"\nAvailable Cell Arrays in {os.path.basename(file_path)}:")
    n_cell_arrays = cell_data.GetNumberOfArrays()
    for i in range(n_cell_arrays):
        print(f"  {i}: {cell_data.GetArrayName(i)}")

    print(f"\nAvailable Point Arrays in {os.path.basename(file_path)}:")
    n_point_arrays = point_data.GetNumberOfArrays()
    for i in range(n_point_arrays):
        print(f"  {i}: {point_data.GetArrayName(i)}")

    # Try to get fill_level from cell data first
    fill_level = cell_data.GetArray("fill_level")
    if fill_level is not None:
        print("Found fill level in cell data with name: fill_level")
        np_array = numpy_support.vtk_to_numpy(fill_level)
        print(f"Array shape before reshape: {np_array.shape}")
        np_array = np_array.reshape(cell_dims[2], cell_dims[1], cell_dims[0])
        print(f"Array shape after reshape: {np_array.shape}")
        return np_array

    raise ValueError("fill_level data not found in the VTI file")

def extract_fill_level_data(_pvd_file):
    """Extract fill level data from all timesteps"""
    timesteps, filenames = read_pvd_timesteps(_pvd_file)
    base_dir = os.path.dirname(_pvd_file)

    # Get domain dimensions from first file
    first_file = os.path.join(base_dir, filenames[0])
    _x_num, _y_num, _z_num = get_domain_dimensions(first_file)
    print(f"Domain dimensions: {_x_num} x {_y_num} x {_z_num}")

    _fill_level_data = []
    for filename in filenames:
        file_path = os.path.join(base_dir, filename)
        print(f"\nReading file: {file_path}")
        try:
            fill_level = read_vti_data(file_path)
            fill_level = fill_level.reshape((_z_num, _y_num, _x_num))
            _fill_level_data.append(fill_level)
        except Exception as e:
            print(f"Error reading {filename}: {str(e)}")
            raise

    return np.array(_fill_level_data), _x_num, _y_num, _z_num

def runRayTracer(x_num, y_num, z_num, fill_levels, current_timestep):

    print("current_timestep", current_timestep)
    # fill_levels = np.clip(fill_levels, 0, 1)
    fill_levels = fill_levels.astype(np.float64)

    # Define the tolerance
    tolerance = 1e-8

    # Check if any value in fill_levels is outside the [0, 1] range
    out_of_range = np.logical_or(fill_levels < -tolerance, fill_levels > 1 + tolerance)
    if np.any(out_of_range):
        out_of_range_values = fill_levels[out_of_range]
        out_of_range_indices = np.where(out_of_range)

        for idx, value in zip(zip(*out_of_range_indices), out_of_range_values):
            print(f"Index: {idx}, Value: {value}")

        # raise ValueError(f"Fill level values must be between 0 and 1 inclusive (with tolerance of {tolerance}).")

    # Check for fill levels within the range [0, 1]
    in_range = np.logical_and(fill_levels > 0, fill_levels < 1)
    in_range_values = fill_levels[in_range]
    in_range_indices = np.where(in_range)

    # print("Fill level values within range (0 to 1):")
    # for idx, value in zip(zip(*in_range_indices), in_range_values):
        # if idx == (np.int64(309),):
            # print(f"Index: {idx}, Value: {value}")

    # Constants
    pi: float = np.pi  # Mathematical constant π
    e: float = 1.60217662e-19  # Elementary charge in Coulombs
    beam_current: float = 3e-3  # Beam current in Amperes
    dwell_time: float = 0.4e-6  # Dwell time in seconds
    A: float = 21.5  # Approximation for Ti6Al4V
    # Ti6Al4V = Ti6Al4V.create_Ti6Al4V(T)
    # A: float = Ti6Al4V.atomic_number
    # print('Ti6Al4V.elements: ', Ti6Al4V.elements)
    # print('Ti6Al4V.composition: ', Ti6Al4V.composition)
    # print('A: ', A)
    dx: float = 2e-6  # Cell size in meters
    dy: float = dx  # Cell size in meters
    dz: float = dx  # Cell size in meters
    threshold: float = 0  # Beam profile threshold
    # The voltage at detector d can be computed using the transimpedance T (in units of V /A ) of a virtual amplifier in the context of the model
    transimpedance: float = 1e5  # Transimpedance in Volts per Ampere
    eV_threshold = 50  # Threshold energy in eV
    threshold_energy = eV_threshold * e  # Threshold energy in joules
    # Define the work distance (height of the electron beam gun)
    # w_d = 12e-3  # Work distance in meters (272 mm)

    # Domain setup (Section 2.1.1)
    # Define the 3D domain for the simulation
    x_min, x_max = -0.5 * x_num * dx, 0.5 * x_num * dx  # -10 mm to 10 mm
    # print('x_min, x_max:', x_min, x_max)
    y_min, y_max = -0.5 * y_num * dy, 0.5 * y_num * dy  # -10 mm to 10 mm
    # print('y_min, y_max:', y_min, y_max)
    z_min, z_max = 0, z_num * dz
    # print('z_min, z_max:', z_min, z_max)
    # x_num = int((x_max - x_min) / dx) + 1
    # y_num = int((y_max - y_min) / dy) + 1
    # z_num = int((z_max - z_min) / dz) + 1
    # print(f"Domain dimensions: x={x_num}, y={y_num}, z={z_num}")
    # domain = np.zeros((z_num, y_num, x_num))  # (z, y , x)
    # print('domain shape: ', np.shape(domain))
    # domain[0:2, :, :] = 1  # Solid cells
    # domain[2, :, :] = 0.5  # Interface layer
    fill_levels_reshaped = np.reshape(fill_levels, (z_num, y_num, x_num))
    fill_levels_reshaped = fill_levels_reshaped.astype(np.float64)
    # print('fill_levels_reshaped:', fill_levels_reshaped)
    out_of_range_fill_levels = np.logical_or(fill_levels_reshaped < -tolerance, fill_levels_reshaped > 1 + tolerance)
    if np.any(out_of_range_fill_levels):
        out_of_range_fill_level_values = fill_levels_reshaped[out_of_range_fill_levels]
        out_of_range_fill_level_indices = np.where(out_of_range_fill_levels)

        for idx, value in zip(zip(*out_of_range_fill_level_indices), out_of_range_fill_level_values):
            print(f"Index: {idx}, Value: {value}")
        # raise ValueError(f"Fill level values must be between 0 and 1 inclusive (with tolerance of {tolerance}).")

    sigma = 1.0
    # Different sigma for each dimension
    # sigma = [1.0, 1.0, 0.5]  # [z, y, x]
    fill_levels_smoothed = ndimage.gaussian_filter(fill_levels_reshaped, sigma=sigma, mode='nearest', radius=1)
    fill_levels_smoothed = np.clip(
        fill_levels_smoothed,
        a_min=np.float64(0.0),
        a_max=np.float64(1.0)
    )

    # Check if values are still in bounds
    if not (np.all(fill_levels_smoothed >= 0) and np.all(fill_levels_smoothed <= 1)):
        raise ValueError("Smoothed fill levels out of bounds!")


    def bilinear_interpolate(_x, _y, z_index, _fill_level):  # to compute _z_interp
        x0, y0 = int(_x), int(_y)
        x1, y1 = x0 + 1, y0 + 1

        fx, fy = x - x0, y - y0

        c00 = _fill_level[z_index, y0, x0]
        c10 = _fill_level[z_index, y0, x1]
        c01 = _fill_level[z_index, y1, x0]
        c11 = _fill_level[z_index, y1, x1]

        return (c00 * (1-fx) * (1-fy) +
                c10 * fx * (1-fy) +
                c01 * (1-fx) * fy +
                c11 * fx * fy)


    def quaternion_interpolate(_v1, _v2, _t):  # to compute _normal_interp
        # Convert vectors to quaternions
        q1 = Ro.from_rotvec(_v1).as_quat()
        q2 = Ro.from_rotvec(_v2).as_quat()
        # q1 = Ro.from_rotvec(np.pi * _v1).as_quat()
        # q2 = Ro.from_rotvec(np.pi * _v2).as_quat()

        # Use Slerp to interpolate between the quaternions
        slerp = Slerp([0, 1], Ro.from_quat([q1, q2]))  # Define Slerp with keyframes
        q_interp = slerp(_t).as_quat()  # Interpolate the quaternion at t

        # Convert the interpolated quaternion back to a direction vector
        return Ro.from_quat(q_interp).apply([0, 0, 1])
        # return Ro.from_quat(q_interp).as_rotvec() / np.pi


    def interpolate_normals(_normal_below: np.ndarray, _normal_above: np.ndarray, t: float) -> np.ndarray:
        """Interpolate between two normals using SLERP"""
        if not (0 <= t <= 1):
            raise ValueError(f"Interpolation parameter t must be in [0,1], got {t}")

        # Ensure normalized vectors
        _normal_below = normalize_vector(_normal_below.astype(np.float64))
        _normal_above = normalize_vector(_normal_above.astype(np.float64))

        # Compute dot product
        dot_product = np.clip(np.dot(_normal_below, _normal_above), -1.0, 1.0)

        # Thresholds for special cases
        PARALLEL_THRESHOLD = 0.9999
        ANTIPARALLEL_THRESHOLD = -0.9999

        # Case 1: Nearly parallel
        if dot_product > PARALLEL_THRESHOLD:
            return normalize_vector((1.0 - t) * _normal_below + t * _normal_above)

        # Case 2: Nearly antiparallel
        if dot_product < ANTIPARALLEL_THRESHOLD:
            perp = np.array([1.0, 0.0, 0.0])
            if abs(np.dot(_normal_below, perp)) > 0.9:
                perp = np.array([0.0, 1.0, 0.0])
            perp = normalize_vector(np.cross(_normal_below, perp))
            angle = np.pi * t
            return normalize_vector(_normal_below * np.cos(angle) + perp * np.sin(angle))

        # Case 3: Standard SLERP
        theta = np.arccos(dot_product)
        sin_theta = np.sin(theta)

        if abs(sin_theta) < 1e-10:
            return normalize_vector(_normal_below)

        scale0 = np.sin((1.0 - t) * theta) / sin_theta
        scale1 = np.sin(t * theta) / sin_theta
        return normalize_vector(scale0 * _normal_below + scale1 * _normal_above)


    def interpolate_fill_level_and_normal(_fill_level, surface_normals, _y, _x):
        """Find interface points and interpolate normals"""
        if surface_normals.ndim != 4:
            raise ValueError(f"Incorrect shape for surface normals: {surface_normals.ndim}")

        # Get fill levels at position
        _column = _fill_level[:, _y, _x]

        # Constants with physical meaning
        INTERFACE_THRESHOLD = 0.5    # Standard VOF interface threshold
        GRADIENT_THRESHOLD = 0.1     # Minimum gradient for interface detection
        TOLERANCE = 1e-6            # Numerical tolerance

        # Scan from top to bottom
        for z in range(len(_column)-2, -1, -1):
            current_fill = _column[z]
            next_fill = _column[z + 1]

            # Check for significant gradient
            gradient = abs(next_fill - current_fill)
            if gradient > GRADIENT_THRESHOLD:
                # Check for interface crossing (VOF method)
                '''crosses_interface = (
                        (current_fill >= INTERFACE_THRESHOLD - TOLERANCE >= next_fill) or
                        (current_fill <= INTERFACE_THRESHOLD + TOLERANCE <= next_fill)
                )'''
                crosses_interface = (
                    (current_fill >= (INTERFACE_THRESHOLD - TOLERANCE) and
                     next_fill <= (INTERFACE_THRESHOLD + TOLERANCE))
                )

                if crosses_interface:
                    print(f"Found interface at z={z}")
                    print(f"fill_level_below: {current_fill}")
                    print(f"fill_level_above: {next_fill}")
                    # Linear interpolation with better numerical stability
                    denominator = next_fill - current_fill
                    if abs(denominator) < TOLERANCE:
                        _z_interp = z + 0.5
                        t_interface = 0.5
                    else:
                        # Calculate interpolation parameter
                        t_interface = (INTERFACE_THRESHOLD - current_fill) / denominator
                        if not (0 <= t_interface <= 1):
                            error_msg = (
                                f"\nInterpolation parameter t_interface={t_interface:.6f} is out of bounds [0,1]\n"
                                f"Debug information:\n"
                                f"  - z-index: {z}\n"
                                f"  - current_fill: {current_fill:.6f}\n"
                                f"  - next_fill: {next_fill:.6f}\n"
                                f"  - denominator: {denominator:.6f}\n"
                                f"  - INTERFACE_THRESHOLD: {INTERFACE_THRESHOLD}\n"
                                f"  - Position (y,x): ({_y}, {_x})\n"
                                f"  - Fill level values around interface:\n"
                            )
                            # Nearby fill levels
                            for i in range(max(0, z-2), min(len(_column), z+3)):
                                error_msg += f"    {i:3d} | {_column[i]:.6f}\n"

                            raise ValueError(error_msg)
                        # t_interface = np.clip(t_interface, 0.0, 1.0)
                        _z_interp = z + t_interface

                    print("_z_interp:", _z_interp)
                    _z_height = (_z_interp + 0.5) * dz
                    print("_z_height:", _z_height)

                    # Get and validate normals
                    _normal_below = surface_normals[z, _y, _x].astype(np.float64)
                    _normal_above = surface_normals[z + 1, _y, _x].astype(np.float64)

                    _normal_interp = interpolate_normals(_normal_below, _normal_above, t_interface)

                    return _z_interp, _z_height, next_fill, current_fill, \
                        _normal_interp, _normal_below, _normal_above

        return None, None, None, None, None, None, None


    def compute_surface_normals_sobel(_fill_levels: np.ndarray) -> np.ndarray:
        """Compute surface normals using Sobel filter"""
        _fill_levels = _fill_levels.astype(np.float64)

        if _fill_levels.ndim != 3:
            raise ValueError("Input must be a 3D array")

        TOLERANCE = {
            'interface': 1e-10,
            'norm': 1e-10
        }

        # Compute gradients without initial inversion
        _grad_x = ndimage.sobel(_fill_levels, axis=2, output=np.float64, mode='nearest')
        _grad_y = ndimage.sobel(_fill_levels, axis=1, output=np.float64, mode='nearest')
        _grad_z = ndimage.sobel(_fill_levels, axis=0, output=np.float64, mode='nearest')

        # Stack gradients
        _n_S = np.stack([_grad_x, _grad_y, _grad_z], axis=-1)

        # Create interface mask
        interface_mask = np.logical_and(
            _fill_levels > TOLERANCE['interface'],
            _fill_levels < 1-TOLERANCE['interface']
        )[..., np.newaxis]

        # Ensure upward-pointing normals (from liquid to gas) after gradient calculation
        mask = _n_S[..., 2] < 0
        _n_S[mask] = -_n_S[mask]

        # Normalize
        norm = np.sqrt(np.sum(_n_S * _n_S, axis=-1, keepdims=True))
        norm = np.maximum(norm, TOLERANCE['norm'])

        # Initialize with default upward normal
        _n_S_normalized = np.zeros_like(_n_S)
        _n_S_normalized[..., 2] = 1.0

        # Apply normalization only at interface
        _n_S_normalized = np.where(interface_mask, _n_S / norm, _n_S_normalized)

        return _n_S_normalized


    # Compute surface normals (Section 2.1.1)
    # n_S, (grad_x, grad_y, grad_z) = compute_surface_normals(domain)
    n_S = compute_surface_normals_sobel(fill_levels_smoothed)
    # print('n_S: ', np.shape(n_S))

    # Beam profile (Section 2.1)
    # Define the Gaussian beam profile
    sigma: float = 300e-6 / (4 * dx)  # Beam profile sigma
    x = np.arange(-np.floor(min(x_num//4, 2*sigma)), 1+np.floor(min((x_num//4), 2*sigma)))
    # print('x:', x)
    y = np.arange(-floor(min(y_num//4, 2*sigma)), 1+floor(min((y_num//4), 2*sigma)))
    # print('y:', y)
    X, Y = np.meshgrid(x, y)
    beam_profile = np.exp(-(X ** 2 + Y ** 2) / (2 * sigma ** 2))
    beam_profile /= np.sum(beam_profile)  # Normalize beam profile
    # print('beam_profile: ', beam_profile)

    beam_size = np.shape(beam_profile)[0] * np.shape(beam_profile)[1]
    # print('Beam size:', beam_size)

    # Detector setup (Section 2.1.1)
    # Define positions of the detectors
    detector_height = 4 * z_max  # Detector height in meters (272 mm)
    detector_spacing_x = 1.5 * dx * x_num  # Detector spacing in meters (127 mm)
    detector_spacing_y = detector_spacing_x  # 1.5 * dy * y_num  # Detector spacing in meters (127 mm)
    detector_diameter = 4 * dx  # Detector diameter in meters (50 mm)

    # Define detector positions relative to the origin
    detector_positions = np.array([
        [0.5 * detector_spacing_x, 0, detector_height],
        [-0.5 * detector_spacing_x, 0, detector_height],
        [0, 0.5 * detector_spacing_y, detector_height],
        [0, -0.5 * detector_spacing_y, detector_height]
    ], dtype=np.float64)
    # print('detector_positions:', detector_positions)


    # Define detector vertices for solid angle calculation
    def get_detector_vertices(detector_position: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Get the vertices of an equilateral triangle detector surface oriented symmetrically.

        Parameters:
        detector_position (array-like): The position of the center of the detector.

        Returns:
        tuple: The vertices of the detector surface.
        """
        # Calculate the side length of the detector
        side_length : float = np.sqrt(3) * (detector_diameter / 2)
        height : float = np.sqrt(3) * (side_length / 2)

        # Determine the orientation based on the detector position
        if np.allclose(detector_position[:2], [0, 0]):
            raise ValueError("Detector position cannot be at the origin (0, 0, 0)")
        elif detector_position[0] > 0:  # Right detector
            _v1 = detector_position + np.array([2 / 3 * height, 0, 0], dtype=np.float64)
            _v2 = detector_position + np.array([-height / 3, side_length / 2, 0], dtype=np.float64)
            _v3 = detector_position + np.array([-height / 3, -side_length / 2, 0], dtype=np.float64)
        elif detector_position[0] < 0:  # Left detector
            _v1 = detector_position + np.array([-2 / 3 * height, 0, 0], dtype=np.float64)
            _v2 = detector_position + np.array([height / 3, -side_length / 2, 0], dtype=np.float64)
            _v3 = detector_position + np.array([height / 3, side_length / 2, 0], dtype=np.float64)
        elif detector_position[1] > 0:  # Back detector
            _v1 = detector_position + np.array([0, 2 / 3 * height, 0], dtype=np.float64)
            _v2 = detector_position + np.array([-side_length / 2, -height / 3, 0], dtype=np.float64)
            _v3 = detector_position + np.array([side_length / 2, -height / 3, 0], dtype=np.float64)
        elif detector_position[1] < 0:  # Front detector
            _v1 = detector_position + np.array([0, -2 / 3 * height, 0], dtype=np.float64)
            _v2 = detector_position + np.array([side_length / 2, height / 3, 0], dtype=np.float64)
            _v3 = detector_position + np.array([-side_length / 2, height / 3, 0], dtype=np.float64)
        else:
            raise ValueError("Invalid detector position. Must be in one of the four expected positions.")

        return _v1, _v2, _v3


    # Compute solid angle using Equation A1 (Appendix A)
    def compute_solid_angle(vector1: np.ndarray, vector2: np.ndarray, vector3: np.ndarray) -> float:
        """
        Compute the solid angle of a 3D triangle.

        Args:
            vector1: The first vector of the triangle.
            vector2: The second vector of the triangle.
            vector3: The third vector of the triangle.

        Returns:
            The solid angle of the triangle.
        """
        # Input validation
        if not isinstance(vector1, np.ndarray) or not isinstance(vector2, np.ndarray) or not isinstance(vector3, np.ndarray):
            raise ValueError("Inputs must be numpy arrays")
        if vector1.shape != (3,) or vector2.shape != (3,) or vector3.shape != (3,):
            raise ValueError("Inputs must be 3D vectors")
        if not np.isfinite(vector1).all() or not np.isfinite(vector2).all() or not np.isfinite(vector3).all():
            raise ValueError("Inputs must be finite")

        # Convert to high precision once
        vectors = [vec.astype(np.float64) for vec in (vector1, vector2, vector3)]
        cross_product = np.cross(vectors[1], vectors[2])
        # Check if the cross product is nearly zero
        if np.linalg.norm(cross_product) < 1e-16:
            raise ValueError("Vectors are nearly coplanar")

        # Compute solid angle
        try:
            # Compute the numerator of the solid angle formula
            numerator: float = np.dot(vectors[0], cross_product)
            # Check if the numerator is nearly zero
            if np.abs(numerator) < 1e-16:
                raise ValueError("Vectors are nearly parallel")

            norms = [np.linalg.norm(v) for v in vectors]

            # Compute the denominator of the solid angle formula
            denominator = (
                    norms[0] * norms[1] * norms[2] +
                    np.dot(vectors[0], vectors[1]) * norms[2] +
                    np.dot(vectors[0], vectors[2]) * norms[1] +
                    np.dot(vectors[1], vectors[2]) * norms[0]
            )

            # Compute the solid angle in radians
            solid_angle = 2 * np.arctan2(numerator, denominator)
            # print("Solid angle:", solid_angle)
            return solid_angle
        except ZeroDivisionError:
            raise ValueError("Error computing solid angle: division by zero")
        except np.linalg.LinAlgError:
            raise ValueError("Error computing solid angle: linear algebra error")


    # BSE coefficient functions (Equations 11 and 10, Section 2.1.2)
    def eta_0(_A: float) -> float:
        """Equation 11 from the paper"""
        _eta_0: float = -0.0254 + 0.016 * _A - 1.86e-4 * _A ** 2 + 8.3e-7 * _A ** 3
        print('eta_0: ', _eta_0)
        return _eta_0


    # BSE Coefficient for single atomic targets, found to hold in the range of 10–100 keV
    def eta(theta: float, _A: float) -> float:
        """Equation 10 from the paper"""
        B: float = 0.89
        _eta: float = B * (eta_0(_A) / B) ** np.cos(theta)
        print('eta: ', _eta)
        return _eta


    '''
    # Compute the weighted average BSE coefficient for the alloy
    def compute_weighted_eta(theta: float, alloy: Alloy) -> float:
        """Equation 12 from the paper"""
        weighted_eta: float = 0
        for element, weight_fraction in zip(alloy.elements, alloy.composition):
            # print('element: ', element)
            # print('weight_fraction: ', weight_fraction)
            _A: float = element.atomic_number
            # print('A: ', _A)
            eta_i: float = eta(theta, _A)
            # print('eta_i: ', eta_i)
            weighted_eta += weight_fraction * eta_i
            # print('weighted_eta: ', weighted_eta)
        return weighted_eta
    '''


    # Correction function C(theta) (Equation 14, Section 2.1.3)
    def C(theta: float) -> float:
        print("theta1:", theta)
        """Equation 14 from the paper"""
        theta = np.rad2deg(theta)  # Ensure angle is in degrees for C function
        _C: float = 1.01 - 4.06 * theta + 1.36e-5 * theta ** 2 - 1.71e-6 * theta ** 3
        print('C: ', _C)
        # Check if C is negative and raise an exception
        if _C < 0:
            raise ValueError(f"C function returned a negative value for theta: {theta}. "
                             f"Please check the input and the C function implementation.")
        return _C


    # Scattering function g_BSE (Equation 13, Section 2.1.3)
    def g_BSE(_theta: float, _alpha: float, _beta: float) -> float:
        print("_theta, _alpha, _beta:", _theta, _alpha, _beta)
        """Equation 13 from the paper"""
        # k: float = np.rad2deg(_theta) / 13  # Correct implementation as per the paper
        # print("k: ", k)
        diffusive_part: float = (eta_0(A) / pi) * np.cos(_alpha)
        print("diffusive_part:", diffusive_part)
        # reflective_part: float = ((eta(_theta, A) - eta_0(A)) / C(_theta)) * ((k + 1) / (2 * pi)) * (np.cos(_beta) ** k)
        # if np.abs(reflective_part) < 1e-10:
            # reflective_part = 0
        # print("reflective_part:", reflective_part)

        # Compute a weighted average BSE coefficient for the alloy
        # reflective_part: float = ((compute_weighted_eta(_theta, Ti6Al4V) - eta_0(A)) / C(_theta)) * ((k + 1) / (2 * pi)) * (np.cos(_beta) ** k)

        # print('g_BSE: ', diffusive_part + reflective_part)
        return diffusive_part  #+ reflective_part


    # Calculate the beam direction vector p_E
    def calculate_p_E(_beam_x, _beam_y):
        print(f"calculate_p_E({_beam_x}, {_beam_y}):")
        # Calculate the vector from the beam gun to the beam location
        beam_vector = np.array([_beam_x * dx, _beam_y * dy, -detector_height], dtype=np.float64)
        return normalize_vector(beam_vector)


    # Calculate phi as the angle between p_E and the negative z-axis
    def calculate_phi(_p_E) -> float:
        z_axis = np.array([0.0, 0.0, -1.0], dtype=np.float64)

        # Check if _p_E is approximately equal to the negative z-axis
        '''if np.allclose(_p_E, z_axis, rtol=1e-8, atol=1e-8):
            _phi_rad = 0.0
        else:
            # Calculate the angle between _p_E and the z-axis using arccos
            _phi_rad = calculate_angle(_p_E, z_axis)'''
        _phi_rad = calculate_angle(_p_E, z_axis)
        print(f"calculate_phi({_p_E}, {_phi_rad}):")
        return _phi_rad


    def set_beam_movement(_x_center, _y_center, _x_offset, _y_offset, _movement_type):
        if _movement_type == 'X':
            if _x_offset >= 0:
                _beam_x_start = _x_center + _x_offset
                _beam_x_end = _x_center - _x_offset - 1
            else:
                _beam_x_start = _x_center + _x_offset
                _beam_x_end = _x_center - _x_offset + 1
            _beam_y_start = _y_center + _y_offset
            _beam_y_end = _y_center + _y_offset + 1
        elif _movement_type == 'Y':
            _beam_x_start = _x_center + _x_offset
            _beam_x_end = _x_center + _x_offset + 1
            if _y_offset >= 0:
                _beam_y_start = _y_center + _y_offset
                _beam_y_end = _y_center - _y_offset - 1
            else:
                _beam_y_start = _y_center + _y_offset
                _beam_y_end = _y_center - _y_offset + 1
        elif _movement_type == 'XY':
            if _x_offset >= 0:
                _beam_x_start = _x_center + _x_offset
                _beam_x_end = _x_center - _x_offset - 1
            else:
                _beam_x_start = _x_center + _x_offset
                _beam_x_end = _x_center - _x_offset + 1
            if _y_offset >= 0:
                _beam_y_start = _y_center + _y_offset
                _beam_y_end = _y_center - _y_offset - 1
            else:
                _beam_y_start = _y_center + _y_offset
                _beam_y_end = _y_center - _y_offset + 1
        else:
            raise ValueError("movement_type must be 'X' or 'Y' or 'XY'")

        return _beam_x_start, _beam_x_end, _beam_y_start, _beam_y_end


    # Initialize arrays to store electron counts
    electrons_per_detector_per_step = np.zeros((x_num, y_num, len(detector_positions)))
    total_electrons_per_step = np.zeros((x_num, y_num))

    # Additional arrays to capture detailed information
    electrons_per_detector_per_beam_profile = np.zeros(
        (beam_profile.shape[0], beam_profile.shape[1], len(detector_positions)))
    total_electrons_per_beam_profile = np.zeros((beam_profile.shape[0], beam_profile.shape[1]))

    # Calculate the center indices
    x_center = x_num // 2
    y_center = y_num // 2
    # print('x_center, y_center:', x_center, y_center)

    x_offset: int = 0
    y_offset: int = 0
    movement_type: str = 'Y'

    # Beam X-Movement
    # beam_x_start, beam_x_end = x_center - x_offset, x_center + x_offset + 1
    # beam_y_start, beam_y_end = y_center + y_offset, y_center + y_offset + 1

    # Beam Y-Movement
    # beam_x_start, beam_x_end = x_center + x_offset, x_center + x_offset + 1
    # beam_y_start, beam_y_end = y_center - y_offset, y_center + y_offset + 1

    # For X or Y Movement
    beam_x_start, beam_x_end, beam_y_start, beam_y_end = set_beam_movement(
        x_center, y_center, x_offset, y_offset, movement_type)

    # Determine the step size based on the direction of movement
    step_x = 1 if beam_x_start <= beam_x_end else -1
    step_y = 1 if beam_y_start <= beam_y_end else -1

    # print(beam_x_start, beam_x_end, beam_y_start, beam_y_end, step_x, step_y)


    z_heights = np.zeros(beam_size)
    fill_level_b = np.zeros(beam_size)
    fill_level_a = np.zeros(beam_size)
    normals = np.zeros((beam_size, 3))
    normals_b = np.zeros((beam_size, 3))
    normals_a = np.zeros((beam_size, 3))
    cell_counter = 0


    # total_electrons = 0
    # electrons_per_det = np.zeros(len(detector_positions))
    # electrons_per_beam_loc = np.zeros((x_num, y_num))
    # electrons_per_beam_loc_per_det = np.zeros((x_num, y_num, len(detector_positions)))
    # electrons_per_beam_prof = np.zeros((beam_profile.shape[0], beam_profile.shape[1]))
    # electrons_per_beam_prof_per_det = np.zeros((beam_profile.shape[0], beam_profile.shape[1], len(detector_positions)))


    def normalize_vector(vector: np.ndarray) -> np.ndarray:
        """Normalize a 3D vector."""
        vector = vector.astype(np.float64)  # Convert to float64
        norm = np.linalg.norm(vector)
        if norm < 1e-18:  # Avoid division by zero
            return vector  # Return original if nearly zero
        return vector / norm  # Normalize


    def calculate_angle(_v1: np.ndarray, _v2: np.ndarray) -> float:
        """Calculate the angle between two normalized vectors."""
        _v1 = normalize_vector(_v1)
        _v2 = normalize_vector(_v2)

        dot_product = np.clip(np.dot(_v1, _v2), -1.0, 1.0)

        return np.arccos(dot_product)  # Return angle in radians


    # Main simulation loop to iterate over the simulation domain
    for beam_x in range(beam_x_start, beam_x_end, step_x):
        for beam_y in range(beam_y_start, beam_y_end, step_y):
            beam_loc = (beam_x, beam_y)
            print(f"Beam center is at domain location: {beam_loc}")

            # Calculate the beam direction vector p_E
            # >> p_E = calculate_p_E(beam_x - x_center, beam_y - y_center)
            # >> print('p_E: ', p_E)

            # Calculate phi as the angle between p_E and the negative z-axis
            # phi_rad = np.arctan2(np.sqrt(dx**2 + dy**2), w_d)
            # >> phi_rad = calculate_phi(p_E)
            # print(f"phi_rad: {phi_rad}")

            # Reset counters for this beam location
            total_electrons: int = 0
            electrons_per_detector = np.zeros(len(detector_positions))

            # Iterate over the beam profile
            for profile_x in range(beam_profile.shape[0]):
            # for profile_x in [beam_profile.shape[0]//2 + 2]:
                for profile_y in range(beam_profile.shape[1]):
                # for profile_y in [beam_profile.shape[1]//2]:
                    print(f"Beam profile shape ({profile_x}, {profile_y})")
                    domain_x = beam_loc[0] - beam_profile.shape[0] // 2 + profile_x
                    domain_y = beam_loc[1] - beam_profile.shape[1] // 2 + profile_y
                    print(f"Domain ({domain_x}, {domain_y})")

                    p_E = calculate_p_E(domain_x - x_center, domain_y - y_center)
                    print('p_E: ', p_E)
                    # phi_rad = calculate_phi(p_E)

                    if (0 <= domain_x < x_num) and (0 <= domain_y < y_num):
                        if beam_profile[profile_x, profile_y] > threshold:
                            print('finding interface_z_indices')
                            # print('interface_z_indices:', interface_z)
                            z_interp, z_height, fill_level_below, fill_level_above, normal_interp, normals_below, normals_above \
                                = interpolate_fill_level_and_normal(fill_levels_smoothed, n_S, domain_y, domain_x)
                            if z_interp is not None:
                                z_heights[cell_counter] = z_height
                                fill_level_b[cell_counter] = fill_level_below
                                fill_level_a[cell_counter] = fill_level_above
                                normals[cell_counter] = normal_interp
                                normals_b[cell_counter] = normals_below
                                normals_a[cell_counter] = normals_above
                                cell_counter += 1
                                print('p_E: ', p_E)
                                n_S_local = normal_interp
                                print('n_S_local:', n_S_local)
                                if not np.allclose(np.linalg.norm(n_S_local), 1.0, atol=1e-5):
                                    raise ValueError(f"n_S_local is not normalized, {np.linalg.norm(n_S_local)}")

                                # if is_parallel_to_z(n_S_local, tolerance):
                                    # print('is_parallel_to_z')
                                    # theta_local: float = phi_rad
                                # else:
                                    # '''theta_local: float = np.arccos(
                                        # np.clip(np.dot(-1 * p_E, n_S_local) / (np.linalg.norm(p_E) * np.linalg.norm(n_S_local)), -1.0, 1.0)
                                    # )'''
                                    # print('is_not_parallel_to_z')
                                theta_local = calculate_angle(-p_E, n_S_local)
                                print('theta_local:', theta_local)
                                # if theta_local < 0.0 or theta_local > 0.011:
                                    # raise ValueError(f"theta_local value should be 0.0, found {theta_local}")
                                    # theta_local: 0.016

                                scattering_point = np.array([
                                    (domain_x - beam_x) * dx,
                                    (domain_y - beam_y) * dy,
                                    z_height
                                ], dtype=np.float64)
                                print(f'scattering_point ({domain_x}, {domain_y}, {z_interp}):', scattering_point)

                                for det_idx, det_pos in enumerate(detector_positions):
                                    # print(f"Detector {det_idx + 1} position: {det_pos}")
                                    det_idx: int
                                    # det_pos: np.ndarray[float]
                                    print(f'det_pos[{det_idx + 1}]: {det_pos}')

                                    # Calculate the vector from the scattering point to the detector
                                    det_vec = det_pos - scattering_point
                                    det_vec = normalize_vector(det_vec)
                                    print(f'det_vec[{det_idx + 1}]: {det_vec}')
                                    if not np.allclose(np.linalg.norm(det_vec), 1.0):
                                        raise ValueError("det_vec is not normalized.")

                                    # Calculate the distance from the scattering point to the detector
                                    # distance = np.linalg.norm(det_pos - scattering_point)
                                    # print(f"Distance to detector {det_idx + 1} ({domain_x}, {domain_y}, {z_interp}): {distance:.9f} m")

                                    # Debugging print statements
                                    '''print(f"Detector {det_idx+1} position: {det_pos}")
                                    print(f"Detector Position Type: {type(det_pos)}")
                                    print(f"Scattering Point: {scattering_point}")
                                    print(f"Detector Vector (normalized): {det_vec}")
                                    print(f"Dot product: {np.dot(det_vec, n_S_local)}")'''

                                    # alpha = np.arccos(np.clip(np.dot(det_vec, n_S_local), -1.0, 1.0))
                                    alpha = calculate_angle(det_vec, n_S_local)
                                    print(f'alpha: {alpha}')

                                    p_E_reflected = p_E - 2.0 * np.dot(p_E, n_S_local) * n_S_local
                                    # p_E_reflected = np.where(np.abs(p_E_reflected) < 1e-8, 0.0, p_E_reflected)
                                    p_E_reflected = normalize_vector(p_E_reflected)
                                    print(f'p_E_reflected: {p_E_reflected}')

                                    # beta = np.arccos(np.clip(np.dot(p_E_reflected, det_vec), -1.0, 1.0))
                                    beta = calculate_angle(p_E_reflected, det_vec)
                                    print(f'beta: {beta}')

                                    # Apply energy threshold
                                    # electron_energy = N_0 * e * g_bse_value
                                    # print('electron_energy: ', electron_energy)

                                    # Calculate the initial energy of the primary electrons (E_0)
                                    V_acc = 60e3  # Accelerating voltage in Volts
                                    E_0 = V_acc * e  # Initial energy of primary electrons in joules

                                    # Calculate the energy of the backscatter electrons
                                    E_BSE = E_0 * eta(theta_local, A)
                                    # Compute a weighted average BSE coefficient for the alloy
                                    # E_BSE = E_0 * compute_weighted_eta(theta_local, Ti6Al4V)
                                    # print('E_BSE: ', E_BSE)

                                    # Bifurcation logic for PE and SE
                                    # if E_BSE > threshold_energy:
                                    # Backscatter electrons (BSE)
                                    # Compute scattering function g_bse_value
                                    g_bse_value = g_BSE(theta_local, alpha, beta)
                                    print(f'g_bse_value: {g_bse_value}')

                                    # Ensure g_bse_value is non-negative
                                    if g_bse_value <= 0:
                                        raise ValueError(f"g_bse_value is {g_bse_value} (zero negative), check calculations.")

                                    # Compute solid angle (Equation A1)
                                    v1, v2, v3 = get_detector_vertices(det_pos)
                                    normal = np.cross(v2 - v1, v3 - v1)
                                    normal = normalize_vector(normal)
                                    '''print(f"Detector {det_idx+1}:")
                                    print(f"  Vertices: {v1}, {v2}, {v3}")
                                    print(f"  Normal vector: {normal}")'''

                                    dOmega = compute_solid_angle(v1 - scattering_point,
                                                                 v2 - scattering_point,
                                                                 v3 - scattering_point)
                                    print(f"Detector {det_idx + 1} solid angle: {dOmega}")

                                    # Compute number of electrons collected by the detector (Equation A2)
                                    # The beam ray stays on each grid point ij for the dwell time t_d and distributes N_0 PE into it
                                    N_0 = beam_current * dwell_time / e  # Equation 3, Section 2.1.1
                                    print("N_0:", N_0)

                                    # There are N1 electrons in the chamber after the 1st scattering event
                                    N_1 = N_0 * eta(theta_local, A)  # Equation 4, Section 2.1.1
                                    # Compute a weighted average BSE coefficient for the alloy
                                    # N_1 = N_0 * compute_weighted_eta(theta_local, Ti6Al4V)  # Equation 4, Section 2.1.1
                                    print("N_1:", N_1)

                                    # The number of electrons N_1_d collected by a general detector d
                                    N_1_d = N_1 * np.sum(g_bse_value * dOmega)  # Equation 5, Section 2.1.1

                                    # total_electrons += N_1_d
                                    # electrons_per_det[det_idx] += N_1_d
                                    # electrons_per_beam_loc[beam_x, beam_y] += N_1_d
                                    # electrons_per_beam_loc_per_det[beam_x, beam_y, det_idx] += N_1_d
                                    # electrons_per_beam_prof[profile_x, profile_y] += N_1_d
                                    # electrons_per_beam_prof_per_det[profile_x, profile_y, det_idx] += N_1_d

                                    # Accumulate electrons per detector for this beam profile point
                                    electrons_per_detector_per_beam_profile[profile_x, profile_y, det_idx] += N_1_d

                                    # Accumulate electrons per detector for this domain point
                                    electrons_per_detector_per_step[beam_x, beam_y, det_idx] += N_1_d

                                    # Add to total electrons for this beam profile point
                                    total_electrons_per_beam_profile[profile_x, profile_y] += N_1_d

                                    # Add to total electrons for this domain point
                                    total_electrons_per_step[beam_x, beam_y] += N_1_d

                                    # Debugging print statements
                                    # print(f"Before addition: electrons_per_detector[{det_idx + 1}]: {electrons_per_detector[det_idx]}")
                                    # print(f"N_0: {N_0}, N_1: {N_1}, g_bse_value: {g_bse_value}, dOmega: {dOmega}")
                                    # print(f"N_1_d: {N_1_d}")

                                    # Accumulate electrons per detector
                                    electrons_per_detector[det_idx] += N_1_d

                                    # Add to total electrons
                                    total_electrons += N_1_d

                                    # print(f"After addition: electrons_per_detector[{det_idx + 1}]: {electrons_per_detector[det_idx]}")
                            else:
                                print(f"No interface found at ({domain_x}, {domain_y})")

                    # Print detailed information for this beam profile point
                    # print(f"Electrons per detector at this beam profile point ({profile_x}, {profile_y}): {electrons_per_detector_per_beam_profile[profile_x, profile_y]}")
                    # print(f"Total electrons at this beam profile point ({profile_x}, {profile_y}): {total_electrons_per_beam_profile[profile_x, profile_y]}")

            # Print detailed information for this domain point after iterating over the beam profile
            # print(f"Electrons per detector at this domain point ({beam_x}, {beam_y}): {electrons_per_detector_per_step[beam_x, beam_y]}")
            # print(f"Total electrons at this domain point ({beam_x}, {beam_y}): {total_electrons_per_step[beam_x, beam_y]}")

            # print("Shape of electrons_per_detector_per_step:", electrons_per_detector_per_step.shape)
            # print("Number of lines:", len(lines))

    # Compute final results
    total_electrons_all_steps = np.sum(total_electrons_per_step)
    total_electrons_per_detector = np.sum(electrons_per_detector_per_step, axis=(0, 1))

    # Print final results
    # print(f"Total electrons hitting all detectors: {total_electrons_all_steps}")
    # for i, electrons in enumerate(total_electrons_per_detector):
        # print(f"Total electrons hitting detector {i + 1}: {electrons}")

    # for beam_x in range(beam_x_start, beam_x_end, step_x):
        # for beam_y in range(beam_y_start, beam_y_end, step_y):
            # print(f"Electrons per detector at this domain point ({beam_x}, {beam_y}): {electrons_per_detector_per_step[beam_x, beam_y]}")
    # print('end simulation')
    # print(electrons_per_detector_per_step)

    # print("total_electrons:", total_electrons, "at timestep", current_timestep)
    # print("electrons_per_det:", electrons_per_det, "at timestep", current_timestep)
    # print("electrons_per_beam_loc:", electrons_per_beam_loc, "at timestep", current_timestep)
    # print("electrons_per_beam_loc_per_det:", electrons_per_beam_loc_per_det, "at timestep", current_timestep)
    return electrons_per_detector_per_step, z_heights, fill_level_b, fill_level_a, normals, normals_b, normals_a

'''
    # Plot results
    # Define the domain points
    domain_points_x = np.arange(beam_x_start, beam_x_end, step_x)
    print(domain_points_x)
    domain_points_y = np.arange(beam_y_start, beam_y_end, step_y)

    # Extract electron counts for each detector based on movement type
    if movement_type == 'X':
        print('movement_type == X')
        detector_1_electrons_x = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, y_center+y_offset, 0].flatten()
        print(detector_1_electrons_x)
        detector_2_electrons_x = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, y_center+y_offset, 1].flatten()
        detector_3_electrons_x = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, y_center+y_offset, 2].flatten()
        detector_4_electrons_x = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, y_center+y_offset, 3].flatten()

        # Plot for detectors 1 and 2
        plt.figure(figsize=(10, 6))
        plt.plot(domain_points_x, detector_1_electrons_x, label='Detector 1', marker='o')
        plt.plot(domain_points_x, detector_2_electrons_x, label='Detector 2', marker='s')
        # Add annotations
        for x, y in zip(domain_points_x, detector_1_electrons_x):
            plt.annotate(f'{y:.2e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
        for x, y in zip(domain_points_x, detector_2_electrons_x):
            plt.annotate(f'{y:.2e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
        plt.title("Electrons Captured by Detectors 1 and 2 (X-axis Movement)")
        plt.xlabel("Domain Point (X-axis)")
        plt.ylabel("Electrons Captured")
        plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.savefig("/local/ca00xebo/repos/walberla/apps/showcases/FreeSurface/plots/detectors_1_2_x_axis.png", dpi=300)
        print('detectors_1_2_x_axis.jpg saved')
        plt.close()

        # Plot for detectors 3 and 4
        plt.figure(figsize=(10, 6))
        plt.plot(domain_points_x, detector_3_electrons_x, label='Detector 3', marker='o')
        plt.plot(domain_points_x, detector_4_electrons_x, label='Detector 4', marker='s')
        # Add annotations
        for x, y in zip(domain_points_x, detector_3_electrons_x):
            plt.annotate(f'{y:.8e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
        for x, y in zip(domain_points_x, detector_4_electrons_x):
            plt.annotate(f'{y:.8e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
        plt.title("Electrons Captured by Detectors 3 and 4 (X-axis Movement)")
        plt.xlabel("Domain Point (X-axis)")
        plt.ylabel("Electrons Captured")
        plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.savefig("detectors_3_4_x_axis.png", dpi=300)
        plt.close()

        # Normalize the data by subtracting the minimum value
        detector_3_normalized = detector_3_electrons_x - np.min(detector_3_electrons_x)
        detector_4_normalized = detector_4_electrons_x - np.min(detector_4_electrons_x)

        plt.figure(figsize=(12, 8))  # Increase figure size for better visibility

        plt.plot(domain_points_x, detector_3_normalized, label='Detector 3', marker='o')
        plt.plot(domain_points_x, detector_4_normalized, label='Detector 4', marker='s')

        # Add annotations for Detector 3
        for x, y in zip(domain_points_x, detector_3_normalized):
            plt.annotate(f'{y:.2f}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
        # Add annotations for Detector 4
        for x, y in zip(domain_points_x, detector_4_normalized):
            plt.annotate(f'{y:.2f}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)

        plt.title("Normalized Electrons Captured by Detectors 3 and 4 (X-axis Movement)")
        plt.xlabel("Domain Point (X-axis)")
        plt.ylabel("Electrons Captured (Normalized)")
        plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.savefig("detectors_3_4_x_axis_normalized.png", dpi=300)
        plt.close()

    elif movement_type == 'Y':
        detector_1_electrons_y = electrons_per_detector_per_step[x_center+x_offset, beam_y_start:beam_y_end:step_y, 0].flatten()
        detector_2_electrons_y = electrons_per_detector_per_step[x_center+x_offset, beam_y_start:beam_y_end:step_y, 1].flatten()
        detector_3_electrons_y = electrons_per_detector_per_step[x_center+x_offset, beam_y_start:beam_y_end:step_y, 2].flatten()
        detector_4_electrons_y = electrons_per_detector_per_step[x_center+x_offset, beam_y_start:beam_y_end:step_y, 3].flatten()

        # Plot for detectors 1 and 2
        plt.figure(figsize=(10, 6))
        plt.plot(domain_points_y, detector_1_electrons_y, label='Detector 1', marker='o')
        plt.plot(domain_points_y, detector_2_electrons_y, label='Detector 2', marker='s')
        # Add annotations
        for x, y in zip(domain_points_y, detector_1_electrons_y):
            plt.annotate(f'{y:.8e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
        for x, y in zip(domain_points_y, detector_2_electrons_y):
            plt.annotate(f'{y:.8e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
        plt.title("Electrons Captured by Detectors 1 and 2 (Y-axis Movement)")
        plt.xlabel("Domain Point (Y-axis)")
        plt.ylabel("Electrons Captured")
        plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.savefig("detectors_1_2_y_axis.png", dpi=300)
        plt.close()

        # Normalize the data by subtracting the minimum value
        detector_1_normalized = detector_1_electrons_y - np.min(detector_1_electrons_y)
        detector_2_normalized = detector_2_electrons_y - np.min(detector_2_electrons_y)

        plt.figure(figsize=(12, 8))  # Increase figure size for better visibility

        plt.plot(domain_points_y, detector_1_normalized, label='Detector 1', marker='o')
        plt.plot(domain_points_y, detector_2_normalized, label='Detector 2', marker='s')

        # Add annotations for Detector 1
        for x, y in zip(domain_points_y, detector_1_normalized):
            plt.annotate(f'{y:.2f}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
        # Add annotations for Detector 2
        for x, y in zip(domain_points_y, detector_2_normalized):
            plt.annotate(f'{y:.2f}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)

        plt.title("Normalized Electrons Captured by Detectors 1 and 2 (Y-axis Movement)")
        plt.xlabel("Domain Point (Y-axis)")
        plt.ylabel("Electrons Captured (Normalized)")
        plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.savefig("detectors_1_2_y_axis_normalized.png", dpi=300)
        plt.close()

        # Plot for detectors 3 and 4
        plt.figure(figsize=(10, 6))
        plt.plot(domain_points_y, detector_3_electrons_y, label='Detector 3', marker='o')
        plt.plot(domain_points_y, detector_4_electrons_y, label='Detector 4', marker='s')
        # Add annotations
        for x, y in zip(domain_points_y, detector_3_electrons_y):
            plt.annotate(f'{y:.8e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
        for x, y in zip(domain_points_y, detector_4_electrons_y):
            plt.annotate(f'{y:.8e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
        plt.title("Electrons Captured by Detectors 3 and 4 (Y-axis Movement)")
        plt.xlabel("Domain Point (Y-axis)")
        plt.ylabel("Electrons Captured")
        plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.savefig("detectors_3_4_y_axis.png", dpi=300)
        plt.close()

    elif movement_type == 'XY':
        # Extract 2D data for XY movement
        detector_1_electrons_xy = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, beam_y_start:beam_y_end:step_y, 0]
        detector_2_electrons_xy = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, beam_y_start:beam_y_end:step_y, 1]
        detector_3_electrons_xy = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, beam_y_start:beam_y_end:step_y, 2]
        detector_4_electrons_xy = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, beam_y_start:beam_y_end:step_y, 3]

        # Create meshgrid for x and y coordinates
        X, Y = np.meshgrid(np.arange(beam_x_start, beam_x_end, step_x), np.arange(beam_y_start, beam_y_end, step_y))

        # Create a figure with multiple subplots
        fig, axs = plt.subplots(2, 2, figsize=(16, 16))

        # Plot for each detector
        for i in range(4):
            row = i // 2
            col = i % 2
            im = axs[row, col].pcolormesh(X, Y, electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, beam_y_start:beam_y_end:step_y, i].T, shading='auto', cmap='viridis', norm=colors.LogNorm())
            axs[row, col].set_title(f"Electrons Captured by Detector {i+1}")
            axs[row, col].set_xlabel("Domain Point (X-axis)")
            axs[row, col].set_ylabel("Domain Point (Y-axis)")
            fig.colorbar(im, ax=axs[row, col])

        # Adjust layout and save
        plt.tight_layout()
        plt.savefig("_all_detectors_xy_movement_.png", dpi=300)
        plt.close()

        # Now let's create individual plots for each detector
        for i in range(4):
            plt.figure(figsize=(10, 8))
            im = plt.pcolormesh(X, Y, electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, beam_y_start:beam_y_end:step_y, i].T, shading='auto', cmap='viridis', norm=colors.LogNorm())
            plt.colorbar(im)
            plt.title(f"Electrons Captured by Detector {i+1}")
            plt.xlabel("Domain Point (X-axis)")
            plt.ylabel("Domain Point (Y-axis)")
            plt.savefig(f"_detector_{i+1}_xy_movement_.png", dpi=300)
            plt.close()

        # --------------------------------------------------

        detector_1_electrons_xy = electrons_per_detector_per_step[beam_x_start:beam_x_end, beam_y_start:beam_y_end, 0]

        # Create the plot
        plt.figure(figsize=(10, 8))
        im = plt.pcolormesh(X, Y, detector_1_electrons_xy, shading='auto', cmap='viridis', norm=colors.LogNorm())
        plt.colorbar(im, label='Electrons Captured')
        plt.title("Electrons Captured by Detector 1 (XY Movement)")
        plt.xlabel("Domain Point (X-axis)")
        plt.ylabel("Domain Point (Y-axis)")

        # Set the correct extent for the plot
        # plt.xlim(beam_x_start, beam_x_end - 1)
        # plt.ylim(beam_y_start, beam_y_end - 1)

        # Adjust color scale to highlight variations
        vmin = np.min(detector_1_electrons_xy)
        vmax = np.max(detector_1_electrons_xy)
        plt.clim(float(vmin), float(vmax))

        # Add text annotations for each cell
        for i in range(detector_1_electrons_xy.shape[0]):
            for j in range(detector_1_electrons_xy.shape[1]):
                plt.text(beam_x_start + i, beam_y_start + j, f'{detector_1_electrons_xy[i, j]:.2f}',
                         ha='center', va='center', fontsize=6, color='white')

        plt.tight_layout()
        plt.savefig("detector_1_xy_movement_corrected.png", dpi=300)
        plt.close()

    # Disable interactive mode
    plt.off()

    # Save the plot to a file
    plt.savefig("final_plot.png", dpi=300)

    # Close all figures
    plt.close('all')
    plt.show()
'''

'''
def plotDetectorResults(electrons_per_detector_per_step, num_steps):
    # Extract the electrons captured by each detector at each time step
    detector_1_electrons = electrons_per_detector_per_step[:, :, 0].sum(axis=1)
    detector_2_electrons = electrons_per_detector_per_step[:, :, 1].sum(axis=1)
    detector_3_electrons = electrons_per_detector_per_step[:, :, 2].sum(axis=1)
    detector_4_electrons = electrons_per_detector_per_step[:, :, 3].sum(axis=1)

    # Create a plot
    plt.figure(figsize=(10, 6))
    plt.plot(range(num_steps), detector_1_electrons, label='Detector 1')
    plt.plot(range(num_steps), detector_2_electrons, label='Detector 2')
    plt.plot(range(num_steps), detector_3_electrons, label='Detector 3')
    plt.plot(range(num_steps), detector_4_electrons, label='Detector 4')

    plt.title("Electrons Captured by Detectors over Time")
    plt.xlabel("Time Step")
    plt.ylabel("Electrons Captured")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("detectors_over_time.png", dpi=300)
    plt.close()
'''

__al__ = ['runRayTracer']


def process_single_timestep(_pvd_file, timestep, filename):
    """Process a single VTI file and run ray tracer"""
    base_dir = os.path.dirname(_pvd_file)
    file_path = os.path.join(base_dir, filename)

    # Read the VTI file
    reader = vtk.vtkXMLImageDataReader()
    reader.SetFileName(file_path)
    reader.Update()
    data = reader.GetOutput()

    # Get dimensions
    dims = data.GetDimensions()
    _x_num, _y_num, _z_num = dims[0]-1, dims[1]-1, dims[2]-1

    # Get fill level data
    cell_data = data.GetCellData()
    fill_level = cell_data.GetArray("fill_level")

    if fill_level is None:
        raise ValueError("fill_level data not found in the VTI file")

    # Convert to numpy array and reshape
    _fill_level_data = numpy_support.vtk_to_numpy(fill_level)
    _fill_level_data = _fill_level_data.reshape((_z_num, _y_num, _x_num))

    # Run ray tracer on this timestep
    print(f"\nProcessing timestep {timestep}")
    electrons_per_detector_per_step, z_heights, fill_level_b, fill_level_a, normals, normals_b, normals_a = runRayTracer(
        _x_num,
        _y_num,
        _z_num,
        _fill_level_data.flatten(),
        timestep
    )

    return electrons_per_detector_per_step, z_heights, fill_level_b, fill_level_a, normals, normals_b, normals_a

def main():
    """Main function to process VTI files and run ray tracer"""
    # Define the path to your pvd file
    # _pvd_file = "/local/ca00xebo/repos/walberla/cmake-build-release/apps/showcases/FreeSurface/vtk_out_11_11_11_2_10/fluid_field.pvd"
    _pvd_file = "/local/ca00xebo/repos/walberla/cmake-build-release/apps/showcases/FreeSurface/vtk_out_51_51_51_12_864/fluid_field.pvd"

    # Read timesteps from pvd file
    timesteps, filenames = read_pvd_timesteps(_pvd_file)
    total_timesteps = len(timesteps)
    electrons_across_timesteps = np.zeros((total_timesteps, 4))

    # Get dimensions from first file to initialize arrays
    first_file = os.path.join(os.path.dirname(_pvd_file), filenames[0])
    reader = vtk.vtkXMLImageDataReader()
    reader.SetFileName(first_file)
    reader.Update()
    dims = reader.GetOutput().GetDimensions()
    x_num, y_num, z_num = dims[0]-1, dims[1]-1, dims[2]-1

    # Process each timestep
    _results = []
    for timestep, filename in enumerate(filenames):
        try:
            print(f"\nProcessing file {filename} (timestep {timestep})")
            result = process_single_timestep(_pvd_file, timestep, filename)
            total_electrons_for_detectors = np.sum(result[0], axis=(0, 1))
            # print("total_electrons_for_detectors:", total_electrons_for_detectors)
            electrons_across_timesteps[timestep] = total_electrons_for_detectors
            # print("electrons_across_timesteps:", electrons_across_timesteps)
            _results.append(result)
            print(f"Completed timestep {timestep}")
        except Exception as e:
            print(f"Error processing timestep {timestep}: {str(e)}")
            raise

    plot_detector_results1(electrons_across_timesteps)

    return _results

def plot_detector_results(electrons_per_detector_time):
    """Plot electron counts for each detector across timesteps"""
    plt.figure(figsize=(12, 8))
    timesteps = range(len(electrons_per_detector_time))

    # Plot for each detector
    for detector in range(4):
        plt.plot(timesteps, electrons_per_detector_time[:, detector],
                 label=f'Detector {detector+1}',
                 marker='o')

    plt.title("Electrons Captured by Detectors Over Time")
    plt.xlabel("Timestep")
    plt.ylabel("Number of Electrons")
    plt.legend()
    plt.grid(True)

    # Add value annotations
    for detector in range(4):
        for t, value in zip(timesteps, electrons_per_detector_time[:, detector]):
            plt.annotate(f'{value:.2e}',
                         (t, value),
                         textcoords="offset points",
                         xytext=(0,10),
                         ha='center',
                         rotation=45)

    plt.tight_layout()
    plt.savefig("detector_electrons_over_time.png", dpi=300, bbox_inches='tight')
    plt.close()

    # Create additional visualization - normalized values
    plt.figure(figsize=(12, 8))
    normalized_electrons = electrons_per_detector_time / np.max(electrons_per_detector_time)

    for detector in range(4):
        plt.plot(timesteps, normalized_electrons[:, detector],
                 label=f'Detector {detector+1}',
                 marker='o')

    plt.title("Normalized Electrons Captured by Detectors Over Time")
    plt.xlabel("Timestep")
    plt.ylabel("Normalized Electron Count")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("detector_electrons_over_time_normalized.png", dpi=300, bbox_inches='tight')
    plt.close()

def plot_detector_results1(electrons_per_detector_time):
    """
    Plot electron counts for detector pairs across timesteps

    Args:
        electrons_per_detector_time: Array of shape (num_timesteps, 4) containing
                                   electron counts for each detector at each timestep
                                   :param electrons_per_detector_time:
    """
    # Create figure for detectors 1 and 2
    plt.figure(figsize=(12, 8))
    timesteps = range(len(electrons_per_detector_time))

    # Plot detectors 1 and 2
    plt.plot(timesteps, electrons_per_detector_time[:, 0],
             label='Detector 1', marker='o', linewidth=1, markevery=20)
    plt.plot(timesteps, electrons_per_detector_time[:, 1],
             label='Detector 2', marker='s', linewidth=1, markevery=20)

    # Add value annotations
    '''for t in timesteps:
        plt.annotate(f'{electrons_per_detector_time[t, 0]:.2e}',
                     (t, electrons_per_detector_time[t, 0]),
                     textcoords="offset points",
                     xytext=(0,10),
                     ha='center',
                     rotation=45)
        plt.annotate(f'{electrons_per_detector_time[t, 1]:.2e}',
                     (t, electrons_per_detector_time[t, 1]),
                     textcoords="offset points",
                     xytext=(0,-15),
                     ha='center',
                     rotation=45)'''

    plt.title("Electrons Captured by Detectors 1 and 2 Over Time")
    plt.xlabel("Timestep")
    plt.ylabel("Number of Electrons")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("detectors_1_2_over_time.png", dpi=300, bbox_inches='tight')
    plt.close()

    # Create figure for detectors 3 and 4
    plt.figure(figsize=(12, 8))

    # Plot detectors 3 and 4
    plt.plot(timesteps, electrons_per_detector_time[:, 2],
             label='Detector 3', marker='o', linewidth=1, markevery=20)
    plt.plot(timesteps, electrons_per_detector_time[:, 3],
             label='Detector 4', marker='s', linewidth=1, markevery=20)

    # Add value annotations
    '''for t in timesteps:
        plt.annotate(f'{electrons_per_detector_time[t, 2]:.2e}',
                     (t, electrons_per_detector_time[t, 2]),
                     textcoords="offset points",
                     xytext=(0,10),
                     ha='center',
                     rotation=45)
        plt.annotate(f'{electrons_per_detector_time[t, 3]:.2e}',
                     (t, electrons_per_detector_time[t, 3]),
                     textcoords="offset points",
                     xytext=(0,-15),
                     ha='center',
                     rotation=45)'''

    plt.title("Electrons Captured by Detectors 3 and 4 Over Time")
    plt.xlabel("Timestep")
    plt.ylabel("Number of Electrons")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("detectors_3_4_over_time.png", dpi=300, bbox_inches='tight')
    plt.close()

if __name__ == "__main__":
    results = main()
