import pyopenvdb as vdb
import numpy as np
import os

def read_specific_vdb(filepath):
    """Read VDB file with proper error handling"""
    try:
        print(f"\nAttempting to read: {filepath}")

        # Check if file exists
        if not os.path.exists(filepath):
            raise FileNotFoundError(f"File not found: {filepath}")

        # Try to read metadata first
        try:
            metadata = vdb.readMetadata(filepath)
            print("\nFile metadata:")
            for key, value in metadata.items():
                print(f"{key}: {value}")
        except Exception as meta_e:
            print(f"Metadata read error: {str(meta_e)}")

        # Try different grid names
        grid_names = ["FillingFrac", "fill_level", "density", "float", "cellfillfrac", "temperature", "velocity"]
        grid = None

        for name in grid_names:
            try:
                print(f"\nTrying to read grid: {name}")
                grid = vdb.read(filepath, name)
                if grid is not None:
                    print(f"Successfully read grid: {name}")
                    print(f"Grid type: {type(grid).__name__}")
                    print(f"Background value: {grid.background}")
                    print(f"Active voxel count: {grid.activeVoxelCount}")
                    break
            except Exception as e:
                print(f"Failed to read grid {name}: {str(e)}")
                continue

        if grid is None:
            raise ValueError("Could not read any grid from file")

        # Get dimensions and create array
        dims = grid.evalActiveVoxelDim()
        print(f"\nGrid dimensions: {dims}")

        fill_array = np.zeros(dims, dtype=np.float64)
        grid.copyToArray(fill_array)
        fill_array = np.transpose(fill_array, (2, 1, 0))

        # Print values at specific position
        x = 169
        y = 220
        print(f"\nFill levels at ({x}, {y}):")
        print("z_index | fill_level")
        print("-" * 20)

        # Get fill levels
        fill_levels = fill_array[:, y, x]
        valid_levels = fill_levels[(fill_levels > 0) & (fill_levels < 1)]

        if len(valid_levels) > 0:
            for z, val in enumerate(fill_levels):
                if 0 <= val <= 1:
                    print(f"{z:7d} | {val:.6f}")
        else:
            print("No fill level found between 0 and 1")

        return fill_array

    except Exception as e:
        print(f"Error reading {filepath}")
        print(f"Error details: {str(e)}")
        print("\nDebugging information:")
        print(f"1. File exists: {os.path.exists(filepath)}")
        print(f"2. File size: {os.path.getsize(filepath) if os.path.exists(filepath) else 'N/A'}")
        print(f"3. OpenVDB module location: {vdb.__file__}")
        print("\nTry examining the file structure:")
        print(f"1. Check grid type: vdb_print -l {filepath}")
        print(f"2. Check metadata: vdb_print -m {filepath}")
        print(f"3. Check full info: vdb_print -v {filepath}")
        return None

# Test specific file
# filepath = "/local/ca00xebo/softwares/KiSSAM/Markl-plate/init_geometry.vdb"
filepath = "/local/ca00xebo/softwares/KiSSAM/Markl-plate/geometry-000003200.vdb"
print(f"\nTesting specific file: {os.path.basename(filepath)}")

# Try to read the file
fill_levels = read_specific_vdb(filepath)

if fill_levels is not None:
    print("\nSuccessfully read fill levels")
    print(f"Array shape: {fill_levels.shape}")
    non_zero = fill_levels[fill_levels > 0]
    if len(non_zero) > 0:
        print(f"Number of non-zero cells: {len(non_zero)}")
        print(f"Min non-zero value: {np.min(non_zero):.6f}")
        print(f"Max value: {np.max(fill_levels):.6f}")
else:
    print("\nFailed to read file")
    print("\nTry these steps:")
    print("1. Check OpenVDB version compatibility")
    print("2. Try using vdb_convert to convert the file")
    print("3. Check if the file is corrupted")
    print("4. Try reading with an older version of OpenVDB")
