import pyopenvdb as vdb
import numpy as np
import os
from pathlib import Path
from scipy import ndimage

def read_specific_vdb(filepath):
    """Read only the FillingFrac grid from VDB file"""
    try:
        # Try to read just the FillingFrac grid
        grid = vdb.read(filepath, "FillingFrac")

        # Get dimensions
        dims = grid.evalActiveVoxelDim()
        print(f"Grid dimensions: {dims}")

        # Create array
        fill_array = np.zeros(dims, dtype=np.float64)

        # Copy data
        grid.copyToArray(fill_array)

        # Transpose to ray tracer's expected order (z, y, x)
        fill_array = np.transpose(fill_array, (2, 1, 0))

        # Create standardized array with 256 z cells
        standardized_array = np.zeros((256, dims[1], dims[0]), dtype=np.float64)

        # Copy data from original array to standardized array
        standardized_array[:fill_array.shape[0], :, :] = fill_array

        # Apply Gaussian smoothing
        sigma = 1.0
        standardized_array_smoothed = ndimage.gaussian_filter(standardized_array, sigma=sigma, mode='nearest', radius=1)

        # Print values
        x = 157
        y = 196
        print(f"\nFill levels at ({x}, {y}):")
        print("z_index | fill_level")
        print("-" * 20)

        # Get fill levels
        fill_levels = standardized_array_smoothed[:, y, x]
        valid_levels = fill_levels[(fill_levels > 0) & (fill_levels < 1)]

        if len(valid_levels) > 0:
            for z, val in enumerate(fill_levels):
                if 0 <= val <= 1:
                    print(f"{z:7d} | {val:.6f}")
        else:
            print("No fill level found between 0 and 1")

        return standardized_array_smoothed

    except Exception as e:
        print(f"Error reading {filepath}")
        print(f"Error details: {str(e)}")
        try:
            metadata = vdb.readMetadata(filepath)
            print("\nFile metadata:")
            for key, value in metadata.items():
                print(f"{key}: {value}")
        except Exception as meta_e:
            print(f"Error reading metadata: {str(meta_e)}")
        return None

def process_vdb_file(filepath):
    """Process a single VDB file and print statistics"""
    print(f"\nProcessing file: {os.path.basename(filepath)}")
    fill_levels = read_specific_vdb(filepath)

    if fill_levels is not None:
        print("\nSuccessfully read fill levels")
        print(f"Array shape: {np.shape(fill_levels)}")

        try:
            # Convert to numpy array if not already
            fill_array = np.asarray(fill_levels)

            # Get non-zero elements
            non_zero = fill_array[fill_array > 0]

            # Get interface cells (0 < fill_level < 1)
            interface_cells = fill_array[(fill_array > 0) & (fill_array < 1)]

            # Print statistics
            if len(non_zero) > 0:
                print("\nNon-zero cells statistics:")
                print(f"Number of non-zero cells: {len(non_zero)}")
                print(f"Min non-zero value: {np.min(non_zero):.6f}")
                print(f"Max value: {np.max(fill_array):.6f}")

            if len(interface_cells) > 0:
                print("\nInterface cells statistics:")
                interface_percentage = (len(interface_cells) * 100.0) / len(non_zero)
                print(f"Number of interface cells: {len(interface_cells)} ({interface_percentage:.3f}%)")
                print(f"Min interface value: {np.min(interface_cells):.6f}")
                print(f"Max interface value: {np.max(interface_cells):.6f}")
                print(f"Mean interface value: {np.mean(interface_cells):.6f}")
            else:
                print("\nNo interface cells found (no fill levels between 0 and 1)")

        except Exception as e:
            print(f"Error processing array: {str(e)}")

        return fill_levels
    return None

def process_directory(directory_path):
    """Process all VDB files in a directory"""
    print(f"Processing directory: {directory_path}")

    # Get all VDB files starting with 'geometry-' and sort them by number
    vdb_files = []
    for f in os.listdir(directory_path):
        if f.startswith('geometry-') and f.endswith('.vdb'):
            try:
                # Extract timestep number from filename
                timestep = int(f.split('-')[1].split('.')[0])
                vdb_files.append((timestep, os.path.join(directory_path, f)))
            except (IndexError, ValueError) as e:
                print(f"Skipping file {f}: {str(e)}")

    # Sort by timestep number
    vdb_files.sort()
    total_files = len(vdb_files)

    # Process each file
    results = {}
    for timestep, filepath in vdb_files:
        print(f"\nProcessing timestep {timestep}")
        fill_levels = process_vdb_file(filepath)
        if fill_levels is not None:
            results[timestep] = fill_levels

    if results:
        print(f"\nSuccessfully processed {len(results)} out of {total_files} files from directory")
    else:
        print(f"\nNo files were successfully processed out of {total_files} total files")
    return results

if __name__ == "__main__":
    # Directory path
    directory = "/local/ca00xebo/softwares/KiSSAM/Markl-plate"
    if directory:
        # Process directory
        print("\n=== Processing directory ===")
        directory_results = process_directory(directory)

    # Specific file path
    specific_file = "/local/ca00xebo/softwares/KiSSAM/Markl-plate/geometry-000000200.vdb"
    # specific_file = "/local/ca00xebo/softwares/KiSSAM/Markl-plate/geometry-000003200.vdb"
    if specific_file:
        # Process specific file
        print("\n=== Processing specific file ===")
        specific_result = process_vdb_file(specific_file)
