import vtk
import numpy as np
from vtkmodules.util import numpy_support
import os

def read_vtk_file(filepath):
    """Read VTK file and check all available fields"""
    try:
        # Create legacy VTK reader
        reader = vtk.vtkDataSetReader()
        reader.SetFileName(filepath)
        reader.Update()

        # Get the data
        data = reader.GetOutput()

        # Get dimensions
        dims = data.GetDimensions()
        print(f"Grid dimensions: {dims}")

        # Print available fields (cell data)
        cell_data = data.GetCellData()
        n_arrays = cell_data.GetNumberOfArrays()
        print("\nAvailable Cell Data Fields:")
        print("-" * 30)

        # Dictionary to store all field data
        field_data = {}

        for i in range(n_arrays):
            array = cell_data.GetArray(i)
            name = array.GetName()
            n_components = array.GetNumberOfComponents()
            n_tuples = array.GetNumberOfTuples()
            print(f"Field {i}: {name}")
            print(f"Number of components: {n_components}")
            print(f"Number of tuples: {n_tuples}")
            print("-" * 30)

            # Convert to numpy array
            field_data[name] = numpy_support.vtk_to_numpy(array)

            # If this is cellfillfrac, print values at specific position
            if name == "cellfillfrac":
                print(f"\nCell fill fraction at (159, 197):")
                print("z_index | fill_level")
                print("-" * 20)

                # Reshape to 3D (cell data dimensions are one less than point dimensions)
                cell_dims = (dims[2]-1, dims[1]-1, dims[0]-1)
                fill_data = field_data[name].reshape(cell_dims)

                # Print non-zero values
                fill_at_pos = fill_data[:, 197, 159]  # Note: using your specified coordinates
                for z, val in enumerate(fill_at_pos):
                    if val > 0:
                        print(f"{z:7d} | {val:.6f}")

        return field_data, dims

    except Exception as e:
        print(f"Error reading {filepath}")
        print(f"Error details: {str(e)}")
        return None, None

# Directory containing VTK files
directory = "/local/ca00xebo/softwares/KiSSAM/Markl-plate"

# Get all VTK files and sort them numerically
def get_number(filename):
    # Extract number from filename (e.g., "Temp600.vtk" -> 600)
    import re
    match = re.search(r'Temp(\d+)\.vtk', filename)
    return int(match.group(1)) if match else 0

vtk_files = []
for f in os.listdir(directory):
    if f.startswith('Temp') and f.endswith('.vtk'):
        try:
            number = get_number(f)
            vtk_files.append((number, os.path.join(directory, f)))
        except (IndexError, ValueError) as e:
            print(f"Skipping file {f}: {str(e)}")

# Sort by timestep number
vtk_files.sort()

# Process each file
for timestep, filepath in vtk_files:
    print(f"\nProcessing timestep {timestep}: {os.path.basename(filepath)}")
    field_data, dims = read_vtk_file(filepath)
    if field_data is not None:
        print("\nSuccessfully read data")
        # Print statistics for cellfillfrac if available
        if "cellfillfrac" in field_data:
            fill_data = field_data["cellfillfrac"]
            non_zero = fill_data[fill_data > 0]
            if len(non_zero) > 0:
                print(f"Number of non-zero cells: {len(non_zero)}")
                print(f"Min non-zero value: {np.min(non_zero):.6f}")
                print(f"Max value: {np.max(fill_data):.6f}")
