import numpy as np
import sympy as sp
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from typing import Tuple
# from src.materials.alloys.Alloy import Alloy
# from src.materials.alloys import Ti6Al4V

T = sp.Symbol('T')

# Constants
pi: float = np.pi  # Mathematical constant π
e: float = 1.60217662e-19  # Elementary charge in Coulombs
beam_current: float = 3e-3  # Beam current in Amperes
dwell_time: float = 0.4e-6  # Dwell time in seconds
A: float = 21.5  # Approximation for Ti6Al4V
# Ti6Al4V = Ti6Al4V.create_Ti6Al4V(T)
# A: float = Ti6Al4V.atomic_number
# print('Ti6Al4V.elements: ', Ti6Al4V.elements)
# print('Ti6Al4V.composition: ', Ti6Al4V.composition)
print('A: ', A)
dx: float = 1000e-6  # Cell size in meters
dy: float = 1000e-6  # Cell size in meters
dz: float = 1000e-6  # Cell size in meters
sigma: float = 3000e-6 / (2 * dx)  # Beam profile sigma
threshold: float = 0.0001  # Beam profile threshold
# The voltage at detector d can be computed using the transimpedance T (in units of V /A ) of a virtual amplifier in the context of the model
transimpedance: float = 1e5  # Transimpedance in Volts per Ampere
eV_threshold = 50  # Threshold energy in eV
threshold_energy = eV_threshold * e  # Threshold energy in joules
print('threshold_energy: ', threshold_energy)
# Define the work distance (height of the electron beam gun)
# w_d = 12e-3  # Work distance in meters (272 mm)

# Domain setup (Section 2.1.1)
# Define the 3D domain for the simulation
x_min, x_max = -40000e-6, 40000e-6  # -40 mm to 40 mm
y_min, y_max = -40000e-6, 40000e-6  # -40 mm to 40 mm
z_min, z_max = -2000e-6, 2000e-6
x_num = int((x_max - x_min) / dx) + 1
y_num = int((y_max - y_min) / dy) + 1
z_num = int((z_max - z_min) / dz) + 1
print(f"Domain dimensions: x={x_num}, y={y_num}, z={z_num}")
domain = np.zeros((z_num, y_num, x_num))  # (z, y , x)
print('domain shape: ', np.shape(domain))
domain[0:2, :, :] = 1  # Solid cells
domain[2, :, :] = 0.5  # Interface layer


def compute_surface_normals(_domain: np.ndarray) -> Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray, np.ndarray], int]:
    """
    Compute surface normals from the domain.

    Args:
        _domain (ndarray): The 3D domain array.

    Returns:
        tuple: A tuple containing the surface normals array and the gradients (grad_x, grad_y, grad_z).
    """
    _grad_x = np.gradient(_domain, axis=2)  # x-direction
    _grad_y = np.gradient(_domain, axis=1)  # y-direction
    _grad_z = np.gradient(_domain, axis=0)  # z-direction
    _n_S = np.stack((_grad_x, _grad_y, _grad_z), axis=-1)  # (z, y, x, 3)
    norm = np.linalg.norm(_n_S, axis=-1)  # (z, y, x)
    interface_mask = (_domain > 0) & (_domain < 1)
    nonzero_mask = norm > 0
    valid_mask = interface_mask & nonzero_mask

    _n_S[valid_mask] /= norm[valid_mask][..., np.newaxis]
    _n_S[valid_mask] *= -1  # Invert the surface normals to point towards the detectors

    if not np.allclose(np.linalg.norm(_n_S[valid_mask], axis=-1), 1.0, atol=1e-8):
        raise ValueError("n_S is not normalized.")

    _interface_layer = np.argmax(np.any(interface_mask, axis=(1, 2)))

    return _n_S, (_grad_x, _grad_y, _grad_z), _interface_layer


# Compute surface normals (Section 2.1.1)
n_S, (grad_x, grad_y, grad_z), interface_layer = compute_surface_normals(domain)
print('np.shape(n_S): ', np.shape(n_S))  # (z, y, x, 3)
print('interface_layer: ', interface_layer)

# Beam profile (Section 2.1)
# Define the Gaussian beam profile
x = np.arange(-2, 3)
y = np.arange(-2, 3)
X, Y = np.meshgrid(x, y)
beam_profile = np.exp(-(X ** 2 + Y ** 2) / (2 * sigma ** 2))
beam_profile /= np.sum(beam_profile)  # Normalize beam profile
print('beam_profile: ', beam_profile)

# Detector setup (Section 2.1.1)
# Define positions of the detectors
detector_height = 272e-3  # Detector height in meters (272 mm)
detector_spacing = 127e-3  # Detector spacing in meters (127 mm)
detector_diameter = 50e-3  # Detector diameter in meters (50 mm)

# Define detector positions relative to the origin
detector_positions = np.array([
    [detector_spacing, 0, detector_height],
    [-detector_spacing, 0, detector_height],
    [0, detector_spacing, detector_height],
    [0, -detector_spacing, detector_height]
])
print('detector_positions:', detector_positions)


# Define detector vertices for solid angle calculation
def get_detector_vertices(detector_position: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Get the vertices of an equilateral triangle detector surface oriented symmetrically.

    Parameters:
    detector_position (array-like): The position of the center of the detector.

    Returns:
    tuple: The vertices of the detector surface.
    """
    # Calculate the side length of the detector
    side_length = np.sqrt(3) * (detector_diameter / 2)
    height = np.sqrt(3) * (side_length / 2)

    # Determine the orientation based on the detector position
    if np.allclose(detector_position[:2], [0, 0]):
        raise ValueError("Detector position cannot be at the origin (0, 0, 0)")
    elif detector_position[0] > 0:  # Right detector
        _v1 = detector_position + np.array([2 / 3 * height, 0, 0])
        _v2 = detector_position + np.array([-height / 3, side_length / 2, 0])
        _v3 = detector_position + np.array([-height / 3, -side_length / 2, 0])
    elif detector_position[0] < 0:  # Left detector
        _v1 = detector_position + np.array([-2 / 3 * height, 0, 0])
        _v2 = detector_position + np.array([height / 3, -side_length / 2, 0])
        _v3 = detector_position + np.array([height / 3, side_length / 2, 0])
    elif detector_position[1] > 0:  # Back detector
        _v1 = detector_position + np.array([0, 2 / 3 * height, 0])
        _v2 = detector_position + np.array([-side_length / 2, -height / 3, 0])
        _v3 = detector_position + np.array([side_length / 2, -height / 3, 0])
    elif detector_position[1] < 0:  # Front detector
        _v1 = detector_position + np.array([0, -2 / 3 * height, 0])
        _v2 = detector_position + np.array([side_length / 2, height / 3, 0])
        _v3 = detector_position + np.array([-side_length / 2, height / 3, 0])
    else:
        raise ValueError("Invalid detector position. Must be in one of the four expected positions.")

    return _v1, _v2, _v3


# Compute solid angle using Equation A1 (Appendix A)
def compute_solid_angle(vector1: np.ndarray, vector2: np.ndarray, vector3: np.ndarray) -> float:
    """
    Compute the solid angle of a 3D triangle.

    Args:
        vector1: The first vector of the triangle.
        vector2: The second vector of the triangle.
        vector3: The third vector of the triangle.

    Returns:
        The solid angle of the triangle.
    """
    # Input validation
    if not isinstance(vector1, np.ndarray) or not isinstance(vector2, np.ndarray) or not isinstance(vector3,
                                                                                                    np.ndarray):
        raise ValueError("Inputs must be numpy arrays")
    if vector1.shape != (3,) or vector2.shape != (3,) or vector3.shape != (3,):
        raise ValueError("Inputs must be 3D vectors")
    if not np.isfinite(vector1).all() or not np.isfinite(vector2).all() or not np.isfinite(vector3).all():
        raise ValueError("Inputs must be finite")

    cross_product = np.cross(vector2, vector3)
    # Check if the cross product is nearly zero
    if np.linalg.norm(cross_product) < 1e-8:
        raise ValueError("Vectors are nearly coplanar")

    # Compute solid angle
    try:
        # Compute the numerator of the solid angle formula
        numerator: float = np.dot(vector1, cross_product)
        # Check if the numerator is nearly zero
        if np.abs(numerator) < 1e-8:
            raise ValueError("Vectors are nearly parallel")

        # Compute the denominator of the solid angle formula
        denominator: float = (
                np.linalg.norm(vector1) * np.linalg.norm(vector2) * np.linalg.norm(vector3)
                + np.dot(vector1, vector2) * np.linalg.norm(vector3)
                + np.dot(vector1, vector3) * np.linalg.norm(vector2)
                + np.dot(vector2, vector3) * np.linalg.norm(vector1)
        )

        # Compute the solid angle in radians
        solid_angle = 2 * np.arctan2(numerator, denominator)
        # print("Solid angle:", solid_angle)
        return solid_angle
    except ZeroDivisionError:
        raise ValueError("Error computing solid angle: division by zero")
    except np.linalg.LinAlgError:
        raise ValueError("Error computing solid angle: linear algebra error")


# BSE coefficient functions (Equations 11 and 10, Section 2.1.2)
def eta_0(_A: float) -> float:
    """Equation 11 from the paper"""
    _eta_0: float = -0.0254 + 0.016 * _A - 1.86e-4 * _A ** 2 + 8.3e-7 * _A ** 3
    # print('eta_0: ', _eta_0)
    return _eta_0


# BSE Coefficient for single atomic targets, found to hold in the range of 10–100 keV
def eta(theta: float, _A: float) -> float:
    """Equation 10 from the paper"""
    B: float = 0.89
    _eta: float = B * (eta_0(_A) / B) ** np.cos(theta)
    # print('eta: ', _eta)
    return _eta


'''
# Compute the weighted average BSE coefficient for the alloy
def compute_weighted_eta(theta: float, alloy: Alloy) -> float:
    """Equation 12 from the paper"""
    weighted_eta: float = 0
    for element, weight_fraction in zip(alloy.elements, alloy.composition):
        # print('element: ', element)
        # print('weight_fraction: ', weight_fraction)
        _A: float = element.atomic_number
        # print('A: ', _A)
        eta_i: float = eta(theta, _A)
        # print('eta_i: ', eta_i)
        weighted_eta += weight_fraction * eta_i
        # print('weighted_eta: ', weighted_eta)
    return weighted_eta
'''


# Correction function C(theta) (Equation 14, Section 2.1.3)
def C(theta: float) -> float:
    """Equation 14 from the paper"""
    theta = np.rad2deg(theta)  # Ensure angle is in degrees for C function
    _C: float = 1.01 - 4.06 * theta + 1.36e-5 * theta ** 2 - 1.71e-6 * theta ** 3
    # print('C: ', _C)
    # Check if C is negative and raise an exception
    if _C < 0:
        raise ValueError(f"C function returned a negative value for theta: {theta}. "
                         f"Please check the input and the C function implementation.")
    return _C


# Scattering function g_BSE (Equation 13, Section 2.1.3)
def g_BSE(_theta: float, _alpha: float, _beta: float) -> float:
    """Equation 13 from the paper"""
    # k: float = np.rad2deg(_theta) / 13  # Correct implementation as per the paper
    diffusive_part: float = (eta_0(A) / pi) * np.cos(_alpha)
    # reflective_part: float = ((eta(_theta, A) - eta_0(A)) / C(_theta)) * ((k + 1) / (2 * pi)) * (np.cos(_beta) ** k)

    # Compute a weighted average BSE coefficient for the alloy
    # reflective_part: float = ((compute_weighted_eta(_theta, Ti6Al4V) - eta_0(A)) / C(_theta)) * ((k + 1) / (2 * pi)) * (np.cos(_beta) ** k)

    # print('g_BSE: ', diffusive_part + reflective_part)
    return diffusive_part  #+ reflective_part


# Calculate the beam direction vector p_E
def calculate_p_E(_beam_x, _beam_y):
    # Calculate the vector from the beam gun to the beam location
    beam_vector = np.array([_beam_x * dx, _beam_y * dy, -detector_height])
    # Normalize the beam vector
    _p_E = beam_vector / np.linalg.norm(beam_vector)
    return _p_E


# Calculate phi as the angle between p_E and the negative z-axis
def calculate_phi(_p_E):
    z_axis = np.array([0, 0, -1])
    dot_product = np.dot(_p_E, z_axis)
    _phi_rad = np.arccos(np.clip(dot_product, -1.0, 1.0))
    return _phi_rad


def set_beam_movement(_x_center, _y_center, _x_offset, _y_offset, _movement_type):
    if _movement_type == 'X':
        if _x_offset >= 0:
            _beam_x_start = _x_center + _x_offset
            _beam_x_end = _x_center - _x_offset - 1
        else:
            _beam_x_start = _x_center + _x_offset
            _beam_x_end = _x_center - _x_offset + 1
        _beam_y_start = _y_center + _y_offset
        _beam_y_end = _y_center + _y_offset + 1
    elif _movement_type == 'Y':
        _beam_x_start = _x_center + _x_offset
        _beam_x_end = _x_center + _x_offset + 1
        if _y_offset >= 0:
            _beam_y_start = _y_center + _y_offset
            _beam_y_end = _y_center - _y_offset - 1
        else:
            _beam_y_start = _y_center + _y_offset
            _beam_y_end = _y_center - _y_offset + 1
    elif _movement_type == 'XY':
        if _x_offset >= 0:
            _beam_x_start = _x_center + _x_offset
            _beam_x_end = _x_center - _x_offset - 1
        else:
            _beam_x_start = _x_center + _x_offset
            _beam_x_end = _x_center - _x_offset + 1
        if _y_offset >= 0:
            _beam_y_start = _y_center + _y_offset
            _beam_y_end = _y_center - _y_offset - 1
        else:
            _beam_y_start = _y_center + _y_offset
            _beam_y_end = _y_center - _y_offset + 1
    else:
        raise ValueError("movement_type must be 'X' or 'Y' or 'XY'")

    return _beam_x_start, _beam_x_end, _beam_y_start, _beam_y_end


# Initialize arrays to store electron counts
electrons_per_detector_per_step = np.zeros((x_num, y_num, len(detector_positions)))
total_electrons_per_step = np.zeros((x_num, y_num))

# Additional arrays to capture detailed information
electrons_per_detector_per_beam_profile = np.zeros(
    (beam_profile.shape[0], beam_profile.shape[1], len(detector_positions)))
total_electrons_per_beam_profile = np.zeros((beam_profile.shape[0], beam_profile.shape[1]))

# Calculate the center indices
x_center = (x_num - 1) // 2
y_center = (y_num - 1) // 2

x_offset: int = -5
y_offset: int = 20
movement_type: str = 'X'

# Beam X-Movement
# beam_x_start, beam_x_end = x_center - x_offset, x_center + x_offset + 1
# beam_y_start, beam_y_end = y_center + y_offset, y_center + y_offset + 1

# Beam Y-Movement
# beam_x_start, beam_x_end = x_center + x_offset, x_center + x_offset + 1
# beam_y_start, beam_y_end = y_center - y_offset, y_center + y_offset + 1

# For X or Y Movement
beam_x_start, beam_x_end, beam_y_start, beam_y_end = set_beam_movement(
    x_center, y_center, x_offset, y_offset, movement_type)

# Determine the step size based on the direction of movement
step_x = 1 if beam_x_start <= beam_x_end else -1
step_y = 1 if beam_y_start <= beam_y_end else -1

print(beam_x_start, beam_x_end, beam_y_start, beam_y_end, step_x, step_y)

# exit()

# Main simulation loop to iterate over the simulation domain
for beam_x in range(beam_x_start, beam_x_end, step_x):
    for beam_y in range(beam_y_start, beam_y_end, step_y):
        beam_loc = (beam_x, beam_y)
        print(f"Beam location: {beam_loc}")

        # Calculate the beam direction vector p_E
        p_E = calculate_p_E(beam_x - x_center, beam_y - y_center)
        # print('p_E: ', p_E)

        # Calculate phi as the angle between p_E and the negative z-axis
        # phi_rad = np.arctan2(np.sqrt(dx**2 + dy**2), w_d)
        phi_rad = calculate_phi(p_E)
        # print(f"phi_rad: {phi_rad}")

        # Reset counters for this beam location
        total_electrons: int = 0
        electrons_per_detector = np.zeros(len(detector_positions))

        # Iterate over the beam profile
        for profile_x in range(beam_profile.shape[0]):
            for profile_y in range(beam_profile.shape[1]):
                print(f"Beam profile shape ({profile_x}, {profile_y}):")
                domain_x = beam_loc[0] - beam_profile.shape[0] // 2 + profile_x
                domain_y = beam_loc[1] - beam_profile.shape[1] // 2 + profile_y
                print(f"Domain ({domain_x}, {domain_y}):")

                if (0 <= domain_x < x_num) and (0 <= domain_y < y_num):
                    if beam_profile[profile_x, profile_y] > threshold:
                        n_S_local = n_S[interface_layer, domain_y, domain_x]  # Interface
                        if not np.allclose(np.linalg.norm(n_S_local), 1.0):
                            raise ValueError("n_S_local is not normalized.")

                        # Debugging print statements
                        '''print(f"grad_x: {grad_x[2, domain_y, domain_x]}")
                        print(f"grad_y: {grad_y[2, domain_y, domain_x]}")
                        print(f"grad_z: {grad_z[2, domain_y, domain_x]}")
                        print(f"n_S_local: {n_S_local}")'''

                        if np.allclose(n_S_local, np.array([0, 0, 1])) or np.allclose(n_S_local, np.array([0, 0, -1])):
                            print('hellooo')
                            theta_local: float = phi_rad
                        else:
                            theta_local: float = np.arccos(
                                np.clip(np.dot(p_E, n_S_local) / (np.linalg.norm(p_E) * np.linalg.norm(n_S_local)),
                                        -1.0, 1.0)
                            )

                        scattering_point = np.array([
                            x_min + domain_x * dx,
                            y_min + domain_y * dy,
                            z_min + interface_layer * dz
                        ])
                        print('scattering_point: ', scattering_point)

                        for det_idx, det_pos in enumerate(detector_positions):
                            print(f"Detector {det_idx + 1} position: {det_pos}")
                            det_idx: int
                            det_pos: np.ndarray[float]

                            # Calculate the vector from the scattering point to the detector
                            det_vec = det_pos - scattering_point
                            det_vec /= np.linalg.norm(det_vec)
                            if not np.allclose(np.linalg.norm(det_vec), 1.0):
                                raise ValueError("det_vec is not normalized.")

                            # Calculate the distance from the scattering point to the detector
                            distance = np.linalg.norm(det_pos - scattering_point)
                            print(f"Distance to detector {det_idx + 1} ({domain_x}, {domain_y}): {distance:.9f} m")

                            # Debugging print statements
                            '''print(f"Detector {det_idx+1} position: {det_pos}")
                            print(f"Detector Position Type: {type(det_pos)}")
                            print(f"Scattering Point: {scattering_point}")
                            print(f"Detector Vector (normalized): {det_vec}")
                            print(f"Dot product: {np.dot(det_vec, n_S_local)}")'''

                            alpha = np.arccos(np.clip(np.dot(det_vec, n_S_local), -1.0, 1.0))

                            p_E_reflected = p_E - 2 * np.dot(p_E, n_S_local) * n_S_local
                            p_E_reflected /= np.linalg.norm(p_E_reflected)

                            beta = np.arccos(np.clip(np.dot(p_E_reflected, det_vec), -1.0, 1.0))

                            # Apply energy threshold
                            # electron_energy = N_0 * e * g_bse_value
                            # print('electron_energy: ', electron_energy)

                            # Calculate the initial energy of the primary electrons (E_0)
                            V_acc = 60e3  # Accelerating voltage in Volts
                            E_0 = V_acc * e  # Initial energy of primary electrons in joules

                            # Calculate the energy of the backscatter electrons
                            E_BSE = E_0 * eta(theta_local, A)
                            # Compute a weighted average BSE coefficient for the alloy
                            # E_BSE = E_0 * compute_weighted_eta(theta_local, Ti6Al4V)
                            # print('E_BSE: ', E_BSE)

                            # Bifurcation logic for PE and SE
                            # if E_BSE > threshold_energy:
                            # Backscatter electrons (BSE)
                            # Compute scattering function g_bse_value
                            g_bse_value = g_BSE(theta_local, alpha, beta)

                            # Ensure g_bse_value is non-negative
                            if g_bse_value < 0:
                                raise ValueError("g_bse_value is negative, check calculations.")

                            # Compute solid angle (Equation A1)
                            v1, v2, v3 = get_detector_vertices(det_pos)
                            normal = np.cross(v2 - v1, v3 - v1)
                            normal /= np.linalg.norm(normal)
                            '''print(f"Detector {det_idx+1}:")
                            print(f"  Vertices: {v1}, {v2}, {v3}")
                            print(f"  Normal vector: {normal}")'''

                            dOmega = compute_solid_angle(v1 - scattering_point, v2 - scattering_point,
                                                         v3 - scattering_point)
                            print(f"Detector {det_idx + 1} solid angle: {dOmega}")

                            # Compute number of electrons collected by the detector (Equation A2)
                            # The beam ray stays on each grid point ij for the dwell time t_d and distributes N_0 PE into it
                            N_0 = beam_current * dwell_time / e  # Equation 3, Section 2.1.1

                            # There are N1 electrons in the chamber after the 1st scattering event
                            N_1 = N_0 * eta(theta_local, A)  # Equation 4, Section 2.1.1
                            # Compute a weighted average BSE coefficient for the alloy
                            # N_1 = N_0 * compute_weighted_eta(theta_local, Ti6Al4V)  # Equation 4, Section 2.1.1

                            # The number of electrons N_1_d collected by a general detector d
                            N_1_d = N_1 * np.sum(g_bse_value * dOmega)  # Equation 5, Section 2.1.1

                            # Accumulate electrons per detector for this beam profile point
                            electrons_per_detector_per_beam_profile[profile_x, profile_y, det_idx] += N_1_d

                            # Accumulate electrons per detector for this domain point
                            electrons_per_detector_per_step[beam_x, beam_y, det_idx] += N_1_d

                            # Add to total electrons for this beam profile point
                            total_electrons_per_beam_profile[profile_x, profile_y] += N_1_d

                            # Add to total electrons for this domain point
                            total_electrons_per_step[beam_x, beam_y] += N_1_d

                            # Debugging print statements
                            print(f"Before addition: electrons_per_detector[{det_idx + 1}]: {electrons_per_detector[det_idx]}")
                            print(f"N_0: {N_0}, N_1: {N_1}, g_bse_value: {g_bse_value}, dOmega: {dOmega}")
                            print(f"N_1_d: {N_1_d}")

                            # Accumulate electrons per detector
                            electrons_per_detector[det_idx] += N_1_d

                            # Add to total electrons
                            total_electrons += N_1_d

                            print(f"After addition: electrons_per_detector[{det_idx + 1}]: {electrons_per_detector[det_idx]}")

                # Print detailed information for this beam profile point
                print(f"Electrons per detector at this beam profile point ({profile_x}, {profile_y}): {electrons_per_detector_per_beam_profile[profile_x, profile_y]}")
                print(f"Total electrons at this beam profile point ({profile_x}, {profile_y}): {total_electrons_per_beam_profile[profile_x, profile_y]}")

        # Print detailed information for this domain point after iterating over the beam profile
        print(f"Electrons per detector at this domain point ({beam_x}, {beam_y}): {electrons_per_detector_per_step[beam_x, beam_y]}")
        print(f"Total electrons at this domain point ({beam_x}, {beam_y}): {total_electrons_per_step[beam_x, beam_y]}")

# Compute final results
total_electrons_all_steps = np.sum(total_electrons_per_step)
total_electrons_per_detector = np.sum(electrons_per_detector_per_step, axis=(0, 1))

# Print final results
print(f"Total electrons hitting all detectors: {total_electrons_all_steps}")
for i, electrons in enumerate(total_electrons_per_detector):
    print(f"Total electrons hitting detector {i + 1}: {electrons}")

for beam_x in range(beam_x_start, beam_x_end, step_x):
    for beam_y in range(beam_y_start, beam_y_end, step_y):
        print(f"Electrons per detector at this domain point ({beam_x}, {beam_y}): {electrons_per_detector_per_step[beam_x, beam_y]}")




# Plot results
# Define the domain points
domain_points_x = np.arange(beam_x_start, beam_x_end, step_x)
domain_points_y = np.arange(beam_y_start, beam_y_end, step_y)

# Extract electron counts for each detector based on movement type
if movement_type == 'X':
    detector_1_electrons_x = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, y_center+y_offset, 0].flatten()
    detector_2_electrons_x = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, y_center+y_offset, 1].flatten()
    detector_3_electrons_x = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, y_center+y_offset, 2].flatten()
    detector_4_electrons_x = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, y_center+y_offset, 3].flatten()

    # Plot for detectors 1 and 2
    plt.figure(figsize=(10, 6))
    plt.plot(domain_points_x, detector_1_electrons_x, label='Detector 1', marker='o')
    plt.plot(domain_points_x, detector_2_electrons_x, label='Detector 2', marker='s')
    # Add annotations
    for x, y in zip(domain_points_x, detector_1_electrons_x):
        plt.annotate(f'{y:.2e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
    for x, y in zip(domain_points_x, detector_2_electrons_x):
        plt.annotate(f'{y:.2e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
    plt.title("Electrons Captured by Detectors 1 and 2 (X-axis Movement)")
    plt.xlabel("Domain Point (X-axis)")
    plt.ylabel("Electrons Captured")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("detectors_1_2_x_axis.png", dpi=300)
    plt.close()

    # Plot for detectors 3 and 4
    plt.figure(figsize=(10, 6))
    plt.plot(domain_points_x, detector_3_electrons_x, label='Detector 3', marker='o')
    plt.plot(domain_points_x, detector_4_electrons_x, label='Detector 4', marker='s')
    # Add annotations
    for x, y in zip(domain_points_x, detector_3_electrons_x):
        plt.annotate(f'{y:.8e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
    for x, y in zip(domain_points_x, detector_4_electrons_x):
        plt.annotate(f'{y:.8e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
    plt.title("Electrons Captured by Detectors 3 and 4 (X-axis Movement)")
    plt.xlabel("Domain Point (X-axis)")
    plt.ylabel("Electrons Captured")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("detectors_3_4_x_axis.png", dpi=300)
    plt.close()

    # Normalize the data by subtracting the minimum value
    detector_3_normalized = detector_3_electrons_x - np.min(detector_3_electrons_x)
    detector_4_normalized = detector_4_electrons_x - np.min(detector_4_electrons_x)

    plt.figure(figsize=(12, 8))  # Increase figure size for better visibility

    plt.plot(domain_points_x, detector_3_normalized, label='Detector 3', marker='o')
    plt.plot(domain_points_x, detector_4_normalized, label='Detector 4', marker='s')

    # Add annotations for Detector 3
    for x, y in zip(domain_points_x, detector_3_normalized):
        plt.annotate(f'{y:.2f}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
    # Add annotations for Detector 4
    for x, y in zip(domain_points_x, detector_4_normalized):
        plt.annotate(f'{y:.2f}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)

    plt.title("Normalized Electrons Captured by Detectors 3 and 4 (X-axis Movement)")
    plt.xlabel("Domain Point (X-axis)")
    plt.ylabel("Electrons Captured (Normalized)")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("detectors_3_4_x_axis_normalized.png", dpi=300)
    plt.close()

elif movement_type == 'Y':
    detector_1_electrons_y = electrons_per_detector_per_step[x_center+x_offset, beam_y_start:beam_y_end:step_y, 0].flatten()
    detector_2_electrons_y = electrons_per_detector_per_step[x_center+x_offset, beam_y_start:beam_y_end:step_y, 1].flatten()
    detector_3_electrons_y = electrons_per_detector_per_step[x_center+x_offset, beam_y_start:beam_y_end:step_y, 2].flatten()
    detector_4_electrons_y = electrons_per_detector_per_step[x_center+x_offset, beam_y_start:beam_y_end:step_y, 3].flatten()

    # Plot for detectors 1 and 2
    plt.figure(figsize=(10, 6))
    plt.plot(domain_points_y, detector_1_electrons_y, label='Detector 1', marker='o')
    plt.plot(domain_points_y, detector_2_electrons_y, label='Detector 2', marker='s')
    # Add annotations
    for x, y in zip(domain_points_y, detector_1_electrons_y):
        plt.annotate(f'{y:.8e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
    for x, y in zip(domain_points_y, detector_2_electrons_y):
        plt.annotate(f'{y:.8e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
    plt.title("Electrons Captured by Detectors 1 and 2 (Y-axis Movement)")
    plt.xlabel("Domain Point (Y-axis)")
    plt.ylabel("Electrons Captured")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("detectors_1_2_y_axis.png", dpi=300)
    plt.close()

    # Normalize the data by subtracting the minimum value
    detector_1_normalized = detector_1_electrons_y - np.min(detector_1_electrons_y)
    detector_2_normalized = detector_2_electrons_y - np.min(detector_2_electrons_y)

    plt.figure(figsize=(12, 8))  # Increase figure size for better visibility

    plt.plot(domain_points_y, detector_1_normalized, label='Detector 1', marker='o')
    plt.plot(domain_points_y, detector_2_normalized, label='Detector 2', marker='s')

    # Add annotations for Detector 1
    for x, y in zip(domain_points_y, detector_1_normalized):
        plt.annotate(f'{y:.2f}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
    # Add annotations for Detector 2
    for x, y in zip(domain_points_y, detector_2_normalized):
        plt.annotate(f'{y:.2f}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)

    plt.title("Normalized Electrons Captured by Detectors 1 and 2 (Y-axis Movement)")
    plt.xlabel("Domain Point (Y-axis)")
    plt.ylabel("Electrons Captured (Normalized)")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("detectors_1_2_y_axis_normalized.png", dpi=300)
    plt.close()

    # Plot for detectors 3 and 4
    plt.figure(figsize=(10, 6))
    plt.plot(domain_points_y, detector_3_electrons_y, label='Detector 3', marker='o')
    plt.plot(domain_points_y, detector_4_electrons_y, label='Detector 4', marker='s')
    # Add annotations
    for x, y in zip(domain_points_y, detector_3_electrons_y):
        plt.annotate(f'{y:.8e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
    for x, y in zip(domain_points_y, detector_4_electrons_y):
        plt.annotate(f'{y:.8e}', (x, y), textcoords="offset points", xytext=(5,-15), ha='center', fontsize=8, rotation=45)
    plt.title("Electrons Captured by Detectors 3 and 4 (Y-axis Movement)")
    plt.xlabel("Domain Point (Y-axis)")
    plt.ylabel("Electrons Captured")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("detectors_3_4_y_axis.png", dpi=300)
    plt.close()

elif movement_type == 'XY':
    # Extract 2D data for XY movement
    detector_1_electrons_xy = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, beam_y_start:beam_y_end:step_y, 0]
    detector_2_electrons_xy = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, beam_y_start:beam_y_end:step_y, 1]
    detector_3_electrons_xy = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, beam_y_start:beam_y_end:step_y, 2]
    detector_4_electrons_xy = electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, beam_y_start:beam_y_end:step_y, 3]

    # Create meshgrid for x and y coordinates
    X, Y = np.meshgrid(np.arange(beam_x_start, beam_x_end, step_x), np.arange(beam_y_start, beam_y_end, step_y))

    # Create a figure with multiple subplots
    fig, axs = plt.subplots(2, 2, figsize=(16, 16))

    # Plot for each detector
    for i in range(4):
        row = i // 2
        col = i % 2
        im = axs[row, col].pcolormesh(X, Y, electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, beam_y_start:beam_y_end:step_y, i].T, shading='auto', cmap='viridis', norm=colors.LogNorm())
        axs[row, col].set_title(f"Electrons Captured by Detector {i+1}")
        axs[row, col].set_xlabel("Domain Point (X-axis)")
        axs[row, col].set_ylabel("Domain Point (Y-axis)")
        fig.colorbar(im, ax=axs[row, col])

    # Adjust layout and save
    plt.tight_layout()
    plt.savefig("_all_detectors_xy_movement_.png", dpi=300)
    plt.close()

    # Now let's create individual plots for each detector
    for i in range(4):
        plt.figure(figsize=(10, 8))
        im = plt.pcolormesh(X, Y, electrons_per_detector_per_step[beam_x_start:beam_x_end:step_x, beam_y_start:beam_y_end:step_y, i].T, shading='auto', cmap='viridis', norm=colors.LogNorm())
        plt.colorbar(im)
        plt.title(f"Electrons Captured by Detector {i+1}")
        plt.xlabel("Domain Point (X-axis)")
        plt.ylabel("Domain Point (Y-axis)")
        plt.savefig(f"_detector_{i+1}_xy_movement_.png", dpi=300)
        plt.close()

    # --------------------------------------------------

    detector_1_electrons_xy = electrons_per_detector_per_step[beam_x_start:beam_x_end, beam_y_start:beam_y_end, 0]

    # Create the plot
    plt.figure(figsize=(10, 8))
    im = plt.pcolormesh(X, Y, detector_1_electrons_xy, shading='auto', cmap='viridis', norm=colors.LogNorm())
    plt.colorbar(im, label='Electrons Captured')
    plt.title("Electrons Captured by Detector 1 (XY Movement)")
    plt.xlabel("Domain Point (X-axis)")
    plt.ylabel("Domain Point (Y-axis)")

    # Set the correct extent for the plot
    # plt.xlim(beam_x_start, beam_x_end - 1)
    # plt.ylim(beam_y_start, beam_y_end - 1)

    # Adjust color scale to highlight variations
    vmin = np.min(detector_1_electrons_xy)
    vmax = np.max(detector_1_electrons_xy)
    plt.clim(float(vmin), float(vmax))

    # Add text annotations for each cell
    for i in range(detector_1_electrons_xy.shape[0]):
        for j in range(detector_1_electrons_xy.shape[1]):
            plt.text(beam_x_start + i, beam_y_start + j, f'{detector_1_electrons_xy[i, j]:.2f}',
                     ha='center', va='center', fontsize=6, color='white')

    plt.tight_layout()
    plt.savefig("detector_1_xy_movement_corrected.png", dpi=300)
    plt.close()
